/* 
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
* 
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "os_interface.h"

#define S3G_TRACK_OK    0
#define S3G_TRACK_FAIL  1

typedef enum
{
    S3G_TYPE_MALLOC_TRACK       = 0x01,
    S3G_TYPE_ALLOC_PAGES_TRACK  = 0x02,
    S3G_TYPE_MAP_PAGES_TRACK    = 0x03,
    S3G_TYPE_MAP_IO_TRACK       = 0x04,
    S3G_TYPE_MEM_TRACK_LAST     = 0x05,

}s3g_mem_track_type;

typedef struct __s3g_mem_track* s3g_mem_track_ptr;

typedef struct __s3g_mem_track
{
   s3g_mem_track_ptr next;
   void*             addr;
   const char*       file;
   unsigned int      size;
   unsigned int      line;
   unsigned int      times;
}s3g_mem_track;

static struct os_mutex *memory_track_lock = NULL;
static s3g_mem_track s3g_memory_tracks[S3G_TYPE_MEM_TRACK_LAST-1] = {{0},};

static inline void s3g_mem_track_add(s3g_mem_track_type type, void* addr, const char* file, int line, unsigned long size)
{
    s3g_mem_track *pMemList  = &s3g_memory_tracks[type - 1];
    s3g_mem_track *pMemTrack = NULL;

    if(addr == NULL)
    {
        return;
    }

    pMemTrack = s3g_malloc_priv(sizeof(s3g_mem_track));

    if(pMemTrack == NULL)
    {
        s3g_error("malloc struct s3g_mem_track failed.\n");
        return;
    }

    if(memory_track_lock != NULL)
    {
        s3g_mutex_lock(memory_track_lock);
    }

    pMemTrack->addr = addr;
    pMemTrack->file = file;
    pMemTrack->line = line;
    pMemTrack->size = size;
    pMemTrack->next = pMemList->next;
    pMemTrack->times = pMemList->times++;
    

    pMemList->next  = pMemTrack;
    pMemList->size++;

    if(memory_track_lock != NULL)
    {
        s3g_mutex_unlock(memory_track_lock);
    }
}

static inline int s3g_mem_track_remove(s3g_mem_track_type type, int check_overflow, void* addr, const char* file, int line)
{
    s3g_mem_track *pMemList  = &s3g_memory_tracks[type - 1];
    s3g_mem_track *pTemp, *pPre;
    int status = S3G_TRACK_OK;

    s3g_mutex_lock(memory_track_lock);

    pPre  = pMemList;
    pTemp = pMemList->next;

    while((pTemp != NULL) && (pTemp->addr != addr))
    {
        pPre  = pTemp;
        pTemp = pTemp->next;
    }

    if(pTemp != NULL)
    {
        if(check_overflow)
        {
            char *head = (char*)pTemp->addr;
            char *tail = (char*)pTemp->addr + pTemp->size - 32;
            int  i;

            for(i=0; i < 32; i++)
            {
                if(head[i] != 0x55 || tail[i] != 0x55)
                {
                    s3g_info("OVERFLOW %s: %d, size = %x, times = %d.\n", 
                             pTemp->file, pTemp->line, pTemp->size, pTemp->times);
                    break;
                }
            }
            s3g_memset(pTemp->addr, 0, pTemp->size);
        }
        pPre->next = pTemp->next;
        s3g_free_priv(pTemp);
      
        pMemList->size--;
    }
    else
    {
        s3g_info("type:%d-memory:0x%p: %s line:%d, no found in track list.\n", pMemList->line, addr, file, line);
        status = S3G_TRACK_FAIL;
    }
    s3g_mutex_unlock(memory_track_lock);

    return status;
}

static inline void s3g_mem_track_list(s3g_mem_track* pMemList, struct os_file *file)
{
    char msg[512];
    s3g_mem_track *pList = pMemList->next;
    
    s3g_info("memory type: %d, leaks number: %d\n", pMemList->line, pMemList->size);

    if(file != NULL)
    {
        s3g_vsnprintf(msg, 512, "memory type: %d, leaks number: %d\n", pMemList->line, pMemList->size);
        s3g_file_write(file, msg, s3g_strlen(msg));
    }

    while(pList)
    {
        s3g_info("%s: %d, size = %u, times = %d.\n", pList->file, pList->line, pList->size, pList->times);
        if(file != NULL)
        {
            s3g_vsnprintf(msg, 512, "%s: %d, size = %u, times = %d.\n", pList->file, pList->line, pList->size, pList->times);
            s3g_file_write(file, msg, s3g_strlen(msg));
        }
        pList = pList->next;
    }

}

void s3g_mem_track_init(void)
{
    s3g_mem_track_type type;
    s3g_mem_track  *header = NULL;

    memory_track_lock = s3g_create_mutex();

    for(type = S3G_TYPE_MALLOC_TRACK; type < S3G_TYPE_MEM_TRACK_LAST; type++)
    {
        header = &s3g_memory_tracks[type-1];
        s3g_memset(header, 0, sizeof(*header));
        header->line = type;
    }
}

void s3g_mem_track_list_result(void)
{
    s3g_mem_track_type type;
    s3g_mem_track  *header = NULL;
    struct os_file *file = NULL;
#if S3G_MEM_TRACK_RESULT_TO_FILE
    file = s3g_file_open(s3g_mem_track_result_path);
#endif

    s3g_info("************leak************\n");

    for(type = S3G_TYPE_MALLOC_TRACK; type < S3G_TYPE_MEM_TRACK_LAST; type++)
    {
        header = &s3g_memory_tracks[type-1];

        if(header->next)
        {
            s3g_mem_track_list(header, file);
        }
    }

#if S3G_MEM_TRACK_RESULT_TO_FILE
    s3g_file_close(file);
#endif

    s3g_destroy_mutex(memory_track_lock);

    memory_track_lock = NULL;
}

void *s3g_malloc_track(unsigned long size, const char *file, unsigned int line)
{
    unsigned long track_size     = size;
    void          *returned_addr = NULL;
    char          *malloc_addr   = NULL;

#if S3G_CHECK_MALLOC_OVERFLOW
    track_size += 2 * 32; 
#endif

    malloc_addr = s3g_malloc_priv(track_size);

#if S3G_CHECK_MALLOC_OVERFLOW
    s3g_memset(malloc_addr, 0x55, 32);
    s3g_memset(malloc_addr + track_size - 32, 0x55, 32);
    returned_addr = malloc_addr + 32;
#else
    returned_addr = malloc_addr;
#endif

    s3g_mem_track_add(S3G_TYPE_MALLOC_TRACK, malloc_addr, file, line, track_size);

    return returned_addr;
}

void *s3g_calloc_track(unsigned long size, const char *file, unsigned int line)
{
    unsigned long track_size     = size;
    void          *returned_addr = NULL;
    char          *malloc_addr   = NULL;

#if S3G_CHECK_MALLOC_OVERFLOW
    track_size += 2 * 32; 
#endif

    malloc_addr = s3g_calloc_priv(track_size);

#if S3G_CHECK_MALLOC_OVERFLOW
    s3g_memset(malloc_addr, 0x55, 32);
    s3g_memset(malloc_addr + track_size - 32, 0x55, 32);
    returned_addr = malloc_addr + 32;
#else
    returned_addr = malloc_addr;
#endif

    s3g_mem_track_add(S3G_TYPE_MALLOC_TRACK, malloc_addr, file, line, track_size);

    return returned_addr;
}

void s3g_free_track(void *addr, const char *file, unsigned int line)
{
    void *malloc_addr   = addr;
    int  overflow_check = 0;
#if S3G_CHECK_MALLOC_OVERFLOW
    overflow_check = 1;
    malloc_addr    = (char*)addr - 32;
#endif
    if(addr == NULL)
    {
        s3g_error("free a NULL pointer: %s, %d.\n", file, line);
        return;
    }

    if(s3g_mem_track_remove(S3G_TYPE_MALLOC_TRACK, overflow_check, malloc_addr, file, line) == S3G_TRACK_OK)
    {
        s3g_free_priv(malloc_addr);
    }
}

struct os_pages_memory * 
s3g_allocate_pages_memory_track(int size, int page_size, alloc_pages_flags_t alloc_flags, const char *file, unsigned int line)
{
    struct os_pages_memory *memory = NULL;

    memory = s3g_allocate_pages_memory_priv(size, page_size, alloc_flags);

    s3g_mem_track_add(S3G_TYPE_ALLOC_PAGES_TRACK, memory, file, line, size);

    return memory;
}

void s3g_free_pages_memory_track(struct os_pages_memory *memory, const char *file, unsigned int line)
{
    if(memory == NULL)
    {
        s3g_error("free a NULL pages memory: %s, %d.\n", file, line);
        return;
    }
    s3g_mem_track_remove(S3G_TYPE_ALLOC_PAGES_TRACK, 0, memory, file, line);
    s3g_free_pages_memory_priv(memory);
}

s3g_vm_area_t *s3g_map_pages_memory_track(void *process_context, s3g_map_argu_t *map,
                                      const char *file, unsigned int line)
{
    s3g_vm_area_t *vm_area = s3g_map_pages_memory_priv(process_context, map);

    s3g_mem_track_add(S3G_TYPE_MAP_PAGES_TRACK, vm_area->virt_addr, file, line, vm_area->size);

    return vm_area;
}

void s3g_unmap_pages_memory_track(s3g_vm_area_t *vm_area,
                                  const char *file, unsigned int line)
{
    if(vm_area->virt_addr == NULL)
    {
        s3g_error("unmap a NULL pages memory: %s, %d.\n", file, line);
        return;
    }
    s3g_mem_track_remove(S3G_TYPE_MAP_PAGES_TRACK, 0, vm_area->virt_addr, file, line);
    s3g_unmap_pages_memory_priv(vm_area);
}

s3g_vm_area_t *s3g_map_io_memory_track(void *process_context, s3g_map_argu_t *map,
                              const char *file, unsigned int line)
{
    s3g_vm_area_t *vm_area = s3g_map_io_memory_priv(process_context, map);

    s3g_mem_track_add(S3G_TYPE_MAP_IO_TRACK, vm_area->virt_addr, file, line, vm_area->size);

    return vm_area;
}

void s3g_unmap_io_memory_track(s3g_vm_area_t *vm_area,
                               const char *file, unsigned int line)
{
    if(vm_area->virt_addr == NULL)
    {
        s3g_error("unmap a NULL io memory: %s, %d.\n", file, line);
        return;
    }
    s3g_mem_track_remove(S3G_TYPE_MAP_IO_TRACK, 0, vm_area->virt_addr, file, line);
    s3g_unmap_io_memory_priv(vm_area);
}

