/*
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
*
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "s3g_driver.h"
#include "s3g_ioctl.h"
#include "s3g_sync.h"
#include "kernel_interface.h"
#include "s3g_buffer_object.h"


static int s3g_buffer_object_release(struct inode *inode, struct file *filp)
{
    s3g_buffer_object_t *bo = filp->private_data;

    filp->private_data = NULL;
    if (bo == NULL)
    {
        return -ENOENT;
    }

    if (bo->allocation)
    {
        krnl_unreference_allocation(bo->adapter, bo->allocation);
    }

    if (bo->read_fence)
    {
        fput(bo->read_fence);
        bo->read_fence = NULL;
    }

    if (bo->write_fence)
    {
        fput(bo->write_fence);
        bo->write_fence = NULL;
    }

    if (bo->lock)
    {
        s3g_destroy_mutex(bo->lock);
        bo->lock = NULL;
    }

    s3g_free(bo);
    filp->private_data = NULL;

    return 0;
}

static void s3g_bo_try_recyle_fence(s3g_buffer_object_t *bo)
{
    s3g_fence_t *fence;
    struct file *read = NULL;
    struct file *write = NULL;

    s3g_mutex_lock(bo->lock);

    if (bo->read_fence != NULL)
    {
        fence = bo->read_fence->private_data;
        if (fence->status == OBJ_SIGNALED)
        {
            read = bo->read_fence;
            bo->read_fence = NULL;
        }
    }
    if (bo->write_fence != NULL)
    {
        fence = bo->write_fence->private_data;
        if (fence->status == OBJ_SIGNALED)
        {
            write = bo->write_fence;
            bo->write_fence = NULL;
        }
    }

    s3g_mutex_unlock(bo->lock);

    if (read != NULL)
    {
        fput(read);
    }

    if (write != NULL)
    {
        fput(write);
    }
}

static long s3g_bo_set_fence(s3g_buffer_object_t *bo, unsigned long arg, struct file **oldp)
{
    int err = 0;
    s3g_bo_fence_t    fence = {0};
    struct file *new, *merged, *old;

    if(s3g_copy_from_user(&fence, (void __user*)arg, sizeof(s3g_bo_fence_t)))
    {
        return -1;
    }

    new = fget(fence.fd);
    if(IS_ERR(new))
    {
        return PTR_ERR(new);
    }

    s3g_bo_try_recyle_fence(bo);

    s3g_mutex_lock(bo->lock);
    if(*oldp)
    {
        old = *oldp;
        *oldp = NULL;

        merged = s3g_sync_fence_merge(old, new);
        if(merged)
        {
            *oldp = merged;
        }
        else
        {
            err = -1;
        }
        s3g_mutex_unlock(bo->lock);

        fput(old);
        fput(new);

        s3g_mutex_lock(bo->lock);
    }
    else
    {
        *oldp = new;
    }
    s3g_mutex_unlock(bo->lock);

    return err;
}

static long s3g_bo_get_fence(s3g_buffer_object_t *bo, unsigned long arg, struct file **oldp)
{
    int err = 0;
    s3g_bo_fence_t    fence = {0};

    s3g_bo_try_recyle_fence(bo);

    if(s3g_copy_from_user(&fence, (void __user*)arg, sizeof(s3g_bo_fence_t)))
    {
        return -1;
    }

    fence.fd = -1;

    s3g_mutex_lock(bo->lock);

    if(*oldp)
    {
        err = get_unused_fd_flags(0);

        if(err >= 0)
        {
            fence.fd = err;
            fd_install(fence.fd, *oldp);
            fget(fence.fd);
        }
    }
    s3g_mutex_unlock(bo->lock);

    if(err >= 0 && s3g_copy_to_user((void __user*)arg, &fence, sizeof(s3g_bo_fence_t)))
    {
        err = -1;
    }

    return err >= 0 ? 0 : err;
}

static long s3g_buffer_object_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    long ret = 0;
    s3g_buffer_object_t* bo = file->private_data;

    switch(cmd)
    {
    case S3G_IOCTL_GET_READ_FENCE:
        ret = s3g_bo_get_fence(bo, arg, &bo->read_fence);
        break;
    case S3G_IOCTL_SET_READ_FENCE:
        ret = s3g_bo_set_fence(bo, arg, &bo->read_fence);
        break;
    case S3G_IOCTL_GET_WRITE_FENCE:
        ret = s3g_bo_get_fence(bo, arg, &bo->write_fence);
        break;
    case S3G_IOCTL_SET_WRITE_FENCE:
        ret = s3g_bo_set_fence(bo, arg, &bo->write_fence);
        break;
    default:
        ret = -EINVAL;
        break;
    }

    return ret;
}

static const struct file_operations s3g_buffer_object_fops = {
    .release    = s3g_buffer_object_release,
    .unlocked_ioctl = s3g_buffer_object_ioctl,
};


int s3g_bo_fd_to_handle(s3g_file_t *priv, s3g_handle_fd_t *hfd)
{
    int ret = 0;
    s3g_buffer_object_t *bo;
    struct file *desc = fget(hfd->fd);

    if (IS_ERR(desc))
    {
        s3g_error("%s, failed to get file desc of fd(0x%x). \n", __func__, hfd->fd);
        return PTR_ERR(desc);
    }

    if (desc->f_op != &s3g_buffer_object_fops)
    {
        ret = -EINVAL;
        goto done;
    }

    bo = desc->private_data;
    hfd->handle = bo->allocation;

done:
    fput(desc);
    return ret;
}

int s3g_bo_handle_to_fd(s3g_file_t *priv, s3g_handle_fd_t *hfd)
{
    int fd;
    s3g_buffer_object_t *bo;
    s3g_card_t *s3g = priv->card;

    bo = s3g_calloc(sizeof(s3g_buffer_object_t));
    if (bo == NULL)
    {
        s3g_error("%s, fail to allocate buffer object\n", __func__);
        return -ENOMEM;
    }

    fd = anon_inode_getfd("s3g_buffer_object", &s3g_buffer_object_fops, bo, O_RDWR);

    if (fd < 0)
    {
        s3g_free(bo);
        return E_FAIL;
    }

    bo->adapter = s3g->adapter;
    bo->lock = s3g_create_mutex();
    bo->allocation = hfd->handle;
    bo->read_fence = NULL;
    bo->write_fence = NULL;

    krnl_reference_allocation(s3g->adapter, bo->allocation);
    hfd->fd = fd;
    return 0;
}

