/* 
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
* 
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "s3g_debugfs.h"

static int s3g_debugfs_node_show(struct seq_file *s, void *unused)
{
    s3g_debugfs_node_t *node = (s3g_debugfs_node_t *)(s->private);
    struct os_seq_file seq_file={0};

    seq_file.seq_file = s;

    s3g_mutex_lock(node->dev->lock);

    switch(node->type)
    {
        case DEBUGFS_NODE_DEVICE:
            krnl_debugfs_dump(&seq_file, node->adapter, node->type, &node->hDevice);
            break;

        case DEBUGFS_NODE_HEAP:
            krnl_debugfs_dump(&seq_file, node->adapter, node->type, &node->id);
            break;

        case DEBUGFS_NODE_INFO:
            krnl_debugfs_dump(&seq_file, node->adapter, node->type, NULL);
            break;

        case DEBUGFS_NODE_MEMTRACK:
            krnl_debugfs_dump(&seq_file, node->adapter, node->type, &node->id);
            break;

        default:
            s3g_info("dump unknow node :%d\n", node->type);
            break;
    }

    s3g_mutex_unlock(node->dev->lock);

    return 0;
}

static int s3g_debugfs_node_open(struct inode *inode, struct file *file)
{
    return single_open(file, s3g_debugfs_node_show, inode->i_private);
}

static const struct file_operations debugfs_node_fops = {
        .open       = s3g_debugfs_node_open,
        .read       = seq_read,
        .llseek     = seq_lseek,
        .release    = single_release,
};

s3g_debugfs_device_t* s3g_debugfs_create_device(s3g_card_t *s3g)
{
    s3g_debugfs_device_t *dev;
    int i=0;

    dev = s3g_calloc(sizeof(struct s3g_debugfs_device));
    if (!dev)
    {
        return NULL;
    }

    dev->s3g        = s3g;
    dev->lock       = s3g_create_mutex();
    INIT_LIST_HEAD(&(dev->node_list));

    dev->debug_root = debugfs_create_dir("s3g", NULL);
    if (!dev->debug_root)
    {
        s3g_error("debugfs: failed to create debugfs root directory.\n");
        return NULL;
    }

    for(i = 0; i<DEBUGFS_HEAP_NUM; i++)
    {
        dev->heap[i].type       = DEBUGFS_NODE_HEAP;
        dev->heap[i].id         = i;
        dev->heap[i].adapter    = dev->s3g->adapter;
        dev->heap[i].dev        = dev;

        s3g_vsprintf(dev->heap[i].name, "heap%d", i);

        dev->heap[i].node_dentry = debugfs_create_file(dev->heap[i].name, 0664, dev->debug_root, &(dev->heap[i]), &debugfs_node_fops);

        if (dev->heap[i].node_dentry == NULL)
        {
            s3g_error("Failed to create debugfs node %s\n", dev->heap[i].name);
            return NULL;
        }
    }

    dev->info.type       = DEBUGFS_NODE_INFO;
    dev->info.adapter    = dev->s3g->adapter;
    dev->info.dev        = dev;

    s3g_vsprintf(dev->heap[i].name, "info");

    dev->info.node_dentry = debugfs_create_file(dev->heap[i].name, 0664, dev->debug_root, &(dev->heap[i]), &debugfs_node_fops);

    if (dev->info.node_dentry == NULL)
    {
        s3g_error("Failed to create debugfs node %s\n", dev->info.name);
        return NULL;
    }

    dev->memtrack.type       = DEBUGFS_NODE_MEMTRACK;
    dev->memtrack.adapter    = dev->s3g->adapter;
    dev->memtrack.dev        = dev;
    dev->memtrack.id        = -1;

    s3g_vsprintf(dev->memtrack.name, "memtrack");

    dev->memtrack.node_dentry = debugfs_create_file(dev->memtrack.name, 0664, dev->debug_root, &(dev->memtrack), &debugfs_node_fops);

    if (dev->memtrack.node_dentry == NULL)
    {
        s3g_error("Failed to create debugfs node %s\n", dev->memtrack.name);
        return NULL;
    }

    return dev;
}

int s3g_debugfs_destroy_device(s3g_debugfs_device_t* dev)
{
    s3g_debugfs_node_t *node1 = NULL;
    s3g_debugfs_node_t *node2 = NULL;

    if(dev == NULL)
    {
        s3g_error("destroy a null dubugfs dev\n");
        return -1;
    }

    if(dev->lock)
    {
        s3g_destroy_mutex(dev->lock);
    }

    if(dev->debug_root)
    {
        debugfs_remove_recursive(dev->debug_root);
    }

    list_for_each_entry_safe(node1, node2, &(dev->node_list), list_item)
    {
        s3g_free(node1);
    }

    s3g_free(dev);

    return 0;
}

int s3g_debugfs_add_node(s3g_debugfs_device_t* dev, int type, int id, unsigned int handle, int create_hint)
{
    s3g_debugfs_node_t *node    = s3g_calloc(sizeof(struct s3g_debugfs_node));
    s3g_debugfs_node_t *node1   = NULL;
    s3g_debugfs_node_t *node2   = NULL;

    //s3g_info("s3g_debugfs_add_node %x, %d, %d, %x\n", dev, type, id, handle);

    if (!node || (type != DEBUGFS_NODE_DEVICE))
    {
        return -1;
    }

    node->type              = type;
    node->id                = id;
    node->adapter           = dev->s3g->adapter;
    node->dev               = dev;
    node->create_hint       = create_hint;
    node->hDevice           = handle;

    s3g_mutex_lock(dev->lock);

    list_for_each_entry_safe(node1, node2, &(dev->node_list), list_item)
    {
        if(node1->id == node->id)
        {
            node->min_id++;
        }
    }

    s3g_vsprintf(node->name, "%d_%d", node->id, node->min_id);

    node->node_dentry = debugfs_create_file(node->name, 0664, dev->debug_root, node, &debugfs_node_fops);

    if (!node->node_dentry)
    {
        s3g_error("Failed to create debugfs node %s\n", node->name);
        s3g_free(node);

        s3g_mutex_unlock(dev->lock);
        return -1;
    }

    list_add_tail(&(node->list_item), &(dev->node_list));

    s3g_mutex_unlock(dev->lock);

    return 0;
}

int s3g_debugfs_remove_node(s3g_debugfs_device_t* dev, unsigned int handle)
{
    s3g_debugfs_node_t *node1 = NULL;
    s3g_debugfs_node_t *node2 = NULL;

    s3g_mutex_lock(dev->lock);

    list_for_each_entry_safe(node1, node2, &(dev->node_list), list_item)
    {
        if(node1->hDevice == handle)
        {
            debugfs_remove_recursive(node1->node_dentry);
            list_del(&(node1->list_item));

            s3g_free(node1);
            break;
        }
    }

    s3g_mutex_unlock(dev->lock);

    return 0;
}

