/* 
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
* 
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "os_interface.h"
#include "s3g_driver.h"
#include "s3g_ioctl.h"
#include "s3g_cec.h"
#include "s3g_debugfs.h"

extern s3g_card_t s3g_cards[S3G_MAX_CARDS];
extern int s3g_run_on_qt;
extern int s3g_freezable_patch;

extern void s3g_dump_clips(s3g_card_t *card);

int s3g_add_wait(s3g_card_t *s3g, s3g_file_t *priv, unsigned int wait_events)
{
    unsigned long flags;

    /* do not add empty wait, and one fd only can have one waiter */
    if((wait_events == 0) || (priv->wait_events != 0))
    {
        s3g_error("fd have wait event:0x%x, can not add new wait event:0x%x\n", priv->wait_events, wait_events);
        return -1;
    }

    INIT_LIST_HEAD(&priv->wait_list);

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    priv->wait_events = wait_events;

    list_add(&priv->wait_list, &s3g->wait_list);

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

    return 0;
}

int s3g_remove_wait(s3g_card_t *s3g, s3g_file_t *priv)
{
    unsigned long flags;

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    list_del(&priv->wait_list);

    priv->wait_events = 0;
    priv->curr_events = 0;

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

    return 0;
}

void s3g_get_event(s3g_card_t *s3g, s3g_file_t *priv, s3g_wait_func_t *wait)
{
    unsigned long flags;

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    wait->events     = priv->curr_events;

    priv->curr_events   = 0;

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);
}

int s3g_post_event(s3g_card_t *s3g, unsigned int events)
{
    struct drm_device *dev = pci_get_drvdata(s3g->pdev);
    s3g_file_t    *priv = NULL;
    unsigned long flags; 

    if (events & S3G_EVENT_VSYNC_0_INTERRUPT)
    {
        ++s3g->v_sync_count[0];
        s3g_get_nsecs(&s3g->v_sync_timestamp[0]);
    }

    if (events & S3G_EVENT_VSYNC_1_INTERRUPT)
    {
        ++s3g->v_sync_count[1];
        s3g_get_nsecs(&s3g->v_sync_timestamp[1]);
    }

    if (events & S3G_EVENT_VSYNC_2_INTERRUPT)
    {
        ++s3g->v_sync_count[2];
        s3g_get_nsecs(&s3g->v_sync_timestamp[2]);
    }

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    list_for_each_entry(priv, &s3g->wait_list, wait_list)
    {
        /* only wakeup when wait event happen */
        if(events & priv->wait_events)
        {
            priv->curr_events |= events;
            wake_up_interruptible(&priv->wait_queue);
        }  
    }

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

    if (events == S3G_EVENT_HOTPLUG)
    {
        drm_sysfs_hotplug_event(dev);
    }

#ifdef NEW_S3GFB
    if(s3g->info != NULL)
    {
        if(!s3g->enter_vt)
        {
            s3gfb_handle_event(s3g, events);
        }
        else
        {
            s3g->s3gfb_pending_events |= events;
        }
    }
#endif  

    return 0;
}

static unsigned int s3g_ipc_type_to_event(unsigned int type, int post)
{
    unsigned int event=0;

    if(type == S3G_HDMI_ON_OFF)
    {
        event = post ? S3G_EVENT_HDMI_ON_OFF : S3G_EVENT_HDMI_ON_OFF_FINISH ;
    }
    else if(type == S3G_MONITORES_ON_OFF)
    {
        event = post ? S3G_EVENT_TV_TURN_ON_OFF_MONITORS : S3G_EVENT_TV_TURN_ON_OFF_MONITORS_FINISH ;
    }
    else if(type == S3G_SET_MODE)
    {
        event = post ? S3G_EVENT_TV_SET_MODE : S3G_EVENT_TV_SET_MODE_FINISH;
    }
    
    return event;
}

int s3g_ipc_operation(s3g_card_t *s3g, s3g_file_t *priv, s3g_ipc_operation_t *ipc_operation)
{
    int                     ret      = 0;
    unsigned long           timeout  = 0;
    long                    wait_ret = 0;

    unsigned int            reply_event = 0, post_event = 0;

    s3g_assert(ipc_operation->type < S3G_IPC_TYPE_COUNT);

    s3g_assert(ipc_operation->size_in < MAX_IPC_OPERATION_DATA*4);
    s3g_assert(ipc_operation->size_out < MAX_IPC_OPERATION_DATA*4);

    switch(ipc_operation->operation)
    {
        case S3G_IPC_GET:
            if(ipc_operation->send_data != NULL)
            {
                ret = s3g_copy_to_user(ipc_operation->send_data, s3g->ipc_data[ipc_operation->type].send_data, s3g->ipc_data[ipc_operation->type].size_in);
            }

            break;

        case S3G_IPC_SEND:
            if(ipc_operation->send_data != NULL)
            {
                ret = s3g_copy_from_user(s3g->ipc_data[ipc_operation->type].send_data, ipc_operation->send_data, ipc_operation->size_in);
                s3g->ipc_data[ipc_operation->type].size_in = ipc_operation->size_in;
            }

            reply_event = s3g_ipc_type_to_event(ipc_operation->type, 0);
            post_event  = s3g_ipc_type_to_event(ipc_operation->type, 1);

            if(s3g_add_wait(s3g, priv, reply_event))
            {
                ret = -1;
                break;
            }

            s3g_post_event(s3g, post_event);

            timeout = msecs_to_jiffies(ipc_operation->timeout);

            s3g_assert(timeout > 0);

            wait_ret = wait_event_interruptible_timeout(priv->wait_queue, reply_event & priv->curr_events, timeout);

            s3g_remove_wait(s3g, priv);

            if(wait_ret > 0 && ipc_operation->reply_data != NULL)
            {
                if(s3g->ipc_data[ipc_operation->type].size_out != ipc_operation->size_out)
                {
                    s3g_error("IPC type:%d need reply size:%d, but actual size:%d, return:0x%x\n", 
                            ipc_operation->type, ipc_operation->size_out, s3g->ipc_data[ipc_operation->type].size_out, wait_ret);
                }

                s3g_assert(s3g->ipc_data[ipc_operation->type].size_out == ipc_operation->size_out);
                ret = s3g_copy_to_user(ipc_operation->reply_data, s3g->ipc_data[ipc_operation->type].reply_data, s3g->ipc_data[ipc_operation->type].size_out);
            }else
            {
                s3g_info("ipc wait 0x%x timeout, wait return 0x%x\n", reply_event, wait_ret);
                ret = -1;
            }
            break;

        case S3G_IPC_REPLY:
            if(ipc_operation->reply_data != NULL)
            {
                ret = s3g_copy_from_user(s3g->ipc_data[ipc_operation->type].reply_data, ipc_operation->reply_data, ipc_operation->size_out);
                s3g->ipc_data[ipc_operation->type].size_out = ipc_operation->size_out;
            }

            reply_event = s3g_ipc_type_to_event(ipc_operation->type, 0);

            s3g_post_event(s3g, reply_event);

            break;
        default:
            s3g_info("no this ipc operation\n"); 
            s3g_assert(0);
    }

    return ret;
}

pgprot_t os_get_pgprot_val(unsigned int *cache_type, pgprot_t old_prot, int io_map)
{
    pgprot_t prot = old_prot;

    if(*cache_type == S3G_MEM_UNCACHED)
    {   
        prot = pgprot_noncached(old_prot);
    }
    else if(*cache_type == S3G_MEM_WRITE_COMBINED)
    {
#ifdef CONFIG_X86
#ifdef CONFIG_X86_PAT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,19,0)
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | cachemode2protval(_PAGE_CACHE_MODE_WC));
#else
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | _PAGE_CACHE_WC);
#endif
#else
        prot = pgprot_noncached(old_prot);

        *cache_type = S3G_MEM_UNCACHED;	
#endif
#else
        prot = pgprot_writecombine(old_prot);
#endif
        
    }
#ifdef S3G_AHB_BUS
    else if(*cache_type == S3G_MEM_WRITE_BACK)
    {   
        // for snoop path, current linux kennel defalut is not WA+WB
        prot = __pgprot((pgprot_val(old_prot)&(~(L_PTE_MT_MASK))) | (L_PTE_MT_WRITEBACK | L_PTE_MT_WRITEALLOC));
    }
#endif
    return prot;
}

#ifndef __frv__
#if 0
int s3g_map_system_ram(struct vm_area_struct* vma, s3g_map_argu_t *map)
{
    unsigned long start = vma->vm_start;

    unsigned long i = 0, pfn;

    unsigned int  cache_type;

    struct os_pages_memory *memory = map->memory;
    
#if LINUX_VERSION_CODE <= 0x02040e
    vma->vm_flags |= VM_LOCKED;
#elif LINUX_VERSION_CODE < 0x030700
    vma->vm_flags |= VM_RESERVED;
#else
    vma->vm_flags |= VM_DONTEXPAND | VM_DONTDUMP;
#endif

    cache_type = map->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);

    for(i = 0; i < memory->page_count; i++)
    {
        pfn = page_to_pfn(memory->pages[i]);

        if(remap_pfn_range(vma, start, pfn, PAGE_SIZE, vma->vm_page_prot))
        {
            return -EAGAIN;
        }

        start += PAGE_SIZE;
    }

    return 0;
}
#endif
int s3g_map_system_io(struct vm_area_struct* vma, s3g_map_argu_t *map)
{
    unsigned int  cache_type = map->flags.cache_type;
#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 1);

    map->flags.cache_type = cache_type;

    if (remap_pfn_range(vma, vma->vm_start,
                vma->vm_pgoff,
                vma->vm_end - vma->vm_start,
                vma->vm_page_prot))
    {
        return -EAGAIN;
    }

    return 0;    
}
#if 1
int s3g_map_system_ram(struct vm_area_struct* vma, s3g_map_argu_t *map)
{
    struct os_pages_memory *memory = map->memory;
    struct os_pmem_area    *area   = NULL;
    struct os_pmem_block   *block  = NULL;
	
    unsigned long start = vma->vm_start;
    unsigned long pfn;
    unsigned int  cache_type, curr_block_offset;
	
    int mapped_size = PAGE_ALIGN(map->size);
    int offset      = _ALIGN_DOWN(map->offset, PAGE_SIZE);
    int area_offset = offset % memory->area_size;
    int range_size, left_size  = 0;
    
#if LINUX_VERSION_CODE <= 0x02040e
    vma->vm_flags |= VM_LOCKED;
#elif LINUX_VERSION_CODE < 0x030700
    vma->vm_flags |= VM_RESERVED;
#else
    vma->vm_flags |= VM_DONTEXPAND | VM_DONTDUMP;
#endif

    cache_type = map->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);
	
    area  = &memory->areas[offset / memory->area_size];
    block = area->block;

    curr_block_offset = area_offset + area->start;

    left_size = mapped_size;

    while(left_size > 0)
    {
        if(curr_block_offset >= block->size)
        {
            curr_block_offset -= block->size;
            block++;

            if(block - memory->blocks >= memory->block_cnt)
            {
                s3g_error("SOMETHING WRONG. block overflow.\n");

                s3g_assert(0);
            }
        }

        if(curr_block_offset >= block->size)
        {
            s3g_error("SOMETHING WRONG. block offset excceed block size.\n");

            s3g_assert(0);
        }

        range_size = (block->size - curr_block_offset) > left_size ? 
                     left_size : 
                     block->size - curr_block_offset;

        pfn  = page_to_pfn(block->pages);
        pfn += curr_block_offset / PAGE_SIZE;

        if(remap_pfn_range(vma, start, pfn, range_size, vma->vm_page_prot))
        {
            return -EAGAIN;
        }

        start             += range_size;
        left_size         -= range_size;
        curr_block_offset += range_size;   
    }

    return 0;
}
#endif
#endif

static irqreturn_t s3g_isr(int irq, void *data)
{
    s3g_card_t *s3g = data;

    if((s3g->isr_func)(s3g->adapter))
    {
        tasklet_schedule(&s3g->tasklet);
    }
    
    return IRQ_HANDLED;
}

void s3g_interrupt_init(s3g_card_t *s3g)
{
    /* request isr */
    krnl_get_isr_func(s3g->adapter, (void**)&s3g->isr_func, (void**)&s3g->isr_bf_func);

    krnl_set_callback_func(s3g->adapter, OS_CALLBACK_POST_EVENT, s3g_post_event, s3g);

    tasklet_init(&s3g->tasklet, (void*)s3g->isr_bf_func, (unsigned long)s3g->adapter);
    tasklet_disable(&s3g->tasklet);
    tasklet_enable(&s3g->tasklet); 

    if(s3g_register_interrupt(s3g, s3g_isr))
    {
        s3g_error("s3g_regster_interrupt failed.\n");
    }
    krnl_enable_interrupt(s3g->adapter);
}

void s3g_interrupt_reinit(s3g_card_t *s3g)
{
    /* request isr */
    krnl_get_isr_func(s3g->adapter, (void**)&s3g->isr_func, (void**)&s3g->isr_bf_func);

    krnl_set_callback_func(s3g->adapter, OS_CALLBACK_POST_EVENT, s3g_post_event, s3g);

    tasklet_enable(&s3g->tasklet); 

    if(s3g_register_interrupt(s3g, s3g_isr))
    {
        s3g_error("s3g_regster_interrupt failed.\n");
    }
}

void s3g_interrupt_deinit(s3g_card_t *s3g)
{
    krnl_disable_interrupt(s3g->adapter);

    s3g_unregister_interrupt(s3g);

    tasklet_disable(&s3g->tasklet);
}

static krnl_import_func_list_t s3g_export =
{
    .udelay                        = s3g_udelay,
    .begin_timer                   = s3g_begin_timer,
    .end_timer                     = s3g_end_timer,
    .do_div                        = s3g_do_div,
    .msleep                        = s3g_msleep,
    .assert                        = s3g_assert,
    .copy_from_user                = s3g_copy_from_user,
    .copy_to_user                  = s3g_copy_to_user,
    .slowbcopy_tobus               = s3g_slowbcopy_tobus,
    .slowbcopy_frombus             = s3g_slowbcopy_frombus,
    .memset                        = s3g_memset,
    .memcpy                        = s3g_memcpy,
    .memcmp_priv                   = s3g_memcmp,
    .byte_copy                     = s3g_byte_copy,
    .strcmp                        = s3g_strcmp,
    .strcpy                        = s3g_strcpy,
    .strncmp                       = s3g_strncmp,
    .strncpy                       = s3g_strncpy,
    .strlen                        = s3g_strlen,
    .secure_read32                 = s3g_secure_read32,
    .read32                        = s3g_read32,
    .read16                        = s3g_read16,
    .read8                         = s3g_read8,
    .secure_write32                = s3g_secure_write32,
    .write32                       = s3g_write32,
    .write16                       = s3g_write16,
    .write8                        = s3g_write8,
    .inb                           = s3g_inb,
    .outb                          = s3g_outb,
    .memsetio                      = s3g_memsetio,
    .memcpy_fromio                 = s3g_memcpy_fromio,
    .memcpy_toio                   = s3g_memcpy_toio,
    .file_open                     = s3g_file_open,
    .file_close                    = s3g_file_close,
    .file_read                     = s3g_file_read,
    .file_write                    = s3g_file_write,
    .vsprintf                      = s3g_vsprintf,
    .vsnprintf                     = s3g_vsnprintf,
    .sscanf                        = s3g_sscanf,
    .printk                        = s3g_printk,
    .cb_printk                     = s3g_cb_printk,
    .seq_printf                    = s3g_seq_printf,
    .find_first_zero_bit           = s3g_find_first_zero_bit,
    .find_next_zero_bit            = s3g_find_next_zero_bit,
    .set_bit                       = s3g_set_bit,
    .clear_bit                     = s3g_clear_bit,
    .getsecs                       = s3g_getsecs,
    .get_nsecs                     = s3g_get_nsecs,
    .create_thread                 = s3g_create_thread,
    .destroy_thread                = s3g_destroy_thread,
    .thread_should_stop            = s3g_thread_should_stop,
    .create_atomic                 = s3g_create_atomic,
    .destroy_atomic                = s3g_destroy_atomic,
    .atomic_read                   = s3g_atomic_read,
    .atomic_inc                    = s3g_atomic_inc,
    .atomic_dec                    = s3g_atomic_dec,
    .create_mutex                  = s3g_create_mutex,
    .destroy_mutex                 = s3g_destroy_mutex,
    .mutex_lock                    = s3g_mutex_lock,
    .mutex_lock_killable           = s3g_mutex_lock_killable,
    .mutex_trylock                 = s3g_mutex_trylock,
    .mutex_unlock                  = s3g_mutex_unlock,
    .create_sema                   = s3g_create_sema,
    .destroy_sema                  = s3g_destroy_sema,
    .down                          = s3g_down,
    .down_trylock                  = s3g_down_trylock,
    .up                            = s3g_up,
    .create_rwsema                 = s3g_create_rwsema,
    .destroy_rwsema                = s3g_destroy_rwsema,
    .down_read                     = s3g_down_read,
    .down_write                    = s3g_down_write,
    .up_read                       = s3g_up_read,
    .up_write                      = s3g_up_write,
    .create_spinlock               = s3g_create_spinlock,
    .destroy_spinlock              = s3g_destroy_spinlock,
    .spin_lock                     = s3g_spin_lock,
    .spin_unlock                   = s3g_spin_unlock,
    .spin_lock_irqsave             = s3g_spin_lock_irqsave,
    .spin_unlock_irqrestore        = s3g_spin_unlock_irqrestore,
    .create_event                  = s3g_create_event,
    .destroy_event                 = s3g_destroy_event,
    .wait_event_thread_safe        = s3g_wait_event_thread_safe,
    .wait_event                    = s3g_wait_event,
    .wake_up_event                 = s3g_wake_up_event,
    .thread_wait                   = s3g_thread_wait,
    .thread_wake_up                = s3g_thread_wake_up,
    .dump_stack                    = s3g_dump_stack,
    .try_to_freeze                 = s3g_try_to_freeze,
    .freezable                     = s3g_freezable,
    .clear_freezable               = s3g_clear_freezable,
    .set_freezable                 = s3g_set_freezable,
    .freezing                      = s3g_freezing,
    .get_current_pid               = s3g_get_current_pid,
    .get_current_tid               = s3g_get_current_tid,
    .flush_cache                   = s3g_flush_cache,
    .inv_cache                     = s3g_inv_cache,
    .get_page_phys_address         = s3g_get_page_phys_address,
    .ioremap                       = s3g_ioremap,
    .iounmap_priv                  = s3g_iounmap,
    .mtrr_add                      = s3g_mtrr_add,
    .mtrr_del                      = s3g_mtrr_del,
    .get_mem_info                  = s3g_get_mem_info,
    .register_shrinker             = s3g_register_shrinker,
    .unregister_shrinker           = s3g_unregister_shrinker,
    .pages_memory_swapout          = s3g_pages_memory_swapout,
    .pages_memory_swapin           = s3g_pages_memory_swapin,
    .release_file_storage          = s3g_release_file_storage,
    .get_bus_config                = s3g_get_bus_config,
    .get_platform_config           = s3g_get_platform_config,
    .get_command_status16          = s3g_get_command_status16,
    .write_command_status16        = s3g_write_command_status16,
    .get_command_status32          = s3g_get_command_status32,
    .write_command_status32        = s3g_write_command_status32,
    .get_rom_start_addr            = s3g_get_rom_start_addr,
    .get_rom_save_addr             = s3g_get_rom_save_addr,
    .write_rom_save_addr           = s3g_write_rom_save_addr,
    .get_bar1                      = s3g_get_bar1,
    //gpio
    .gpio_set                      = s3g_gpio_setvalue,
    .gpio_get                      = s3g_gpio_getvalue,
    .gpio_request                  = s3g_gpio_request,
    .gpio_free                     = s3g_gpio_free,
    .gpio_direction_input          = s3g_gpio_direction_input,
    .gpio_direction_output         = s3g_gpio_direction_output,
    //regulator
    .regulator_get                 = s3g_regulator_get,
    .regulator_enable              = s3g_regulator_enable,
    .regulator_disable             = s3g_regulator_disable,
    .regulator_is_enabled          = s3g_regulator_is_enabled,
    .regulator_get_voltage         = s3g_regulator_get_voltage,
    .regulator_set_voltage         = s3g_regulator_set_voltage,
    .regulator_put                 = s3g_regulator_put,

    .malloc_track                  = s3g_malloc_track,
    .calloc_track                  = s3g_calloc_track,
    .free_track                    = s3g_free_track,
    .malloc_priv                   = s3g_malloc_priv,
    .calloc_priv                   = s3g_calloc_priv,
    .free_priv                     = s3g_free_priv,
    .allocate_pages_memory_track   = s3g_allocate_pages_memory_track,
    .free_pages_memory_track       = s3g_free_pages_memory_track,
    .allocate_pages_memory_priv    = s3g_allocate_pages_memory_priv,
    .free_pages_memory_priv        = s3g_free_pages_memory_priv,
    .map_pages_memory_track        = s3g_map_pages_memory_track,
    .unmap_pages_memory_track      = s3g_unmap_pages_memory_track,
    .map_pages_memory_priv         = s3g_map_pages_memory_priv,
    .unmap_pages_memory_priv       = s3g_unmap_pages_memory_priv,
    .map_io_memory_track           = s3g_map_io_memory_track,
    .unmap_io_memory_track         = s3g_unmap_io_memory_track,
    .map_io_memory_priv            = s3g_map_io_memory_priv,
    .unmap_io_memory_priv          = s3g_unmap_io_memory_priv,
    .acquire_pfns                  = s3g_acquire_pfns,
    .release_pfns                  = s3g_release_pfns,
#ifdef S3G_TRACE_EVENT
    .register_trace_events         = s3g_register_trace_events,
    .unregister_trace_events       = s3g_unregister_trace_events,
    .fence_back_trace_event        = s3g_fence_back_trace_event,
    .cmd_render_trace_event        = s3g_cmd_render_trace_event,
    .dma_submit_trace_event        = s3g_dma_submit_trace_event,
#endif

#ifdef ENABLE_TS
    .register_gfx_ts_handle        = s3g_register_gfx_ts_handle,
#else
    .register_gfx_ts_handle        = NULL,
#endif

    .query_platform_caps           = s3g_query_platform_caps,
};

/***/
#include <linux/proc_fs.h>

#ifdef NO_PROC_CREATE_FUNC
static int s3g_read_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{

    s3g_card_t    *s3g = data;

    krnl_dump_resource(s3g->adapter, 5, 0);

    return 0;
}
#else
static ssize_t s3g_proc_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
    s3g_card_t *s3g = PDE_DATA(file_inode(filp));

    krnl_dump_resource(s3g->adapter, 1, 0);

    return 0;
}

#define S3G_DUMP_CLIPS_INDEX           0x300

#define S3G_DUMP_POWER_STATUS_INDEX    0x1000
#define S3G_DUMP_SENSOR_TEMP_INDEX     0x2000
#define S3G_FORCE_DVFS_INDEX           0x3000
#define S3G_FORCE_DVFS_VIDEO_INDEX     0x4000

#define S3G_DUMP_COMPRESS_PS_INDEX    (0x100 + S3G_STREAM_PS)
#define S3G_DUMP_COMPRESS_SS_INDEX    (0x100 + S3G_STREAM_SS)
#define S3G_DUMP_COMPRESS_TS_INDEX    (0x100 + S3G_STREAM_TS)
#define S3G_DUMP_COMPRESS_4S_INDEX    (0x100 + S3G_STREAM_4S)
#define S3G_DUMP_COMPRESS_5S_INDEX    (0x100 + S3G_STREAM_5S)
#define S3G_DUMP_COMPRESS_6S_INDEX    (0x100 + S3G_STREAM_6S)
#define S3G_DUMP_COMPRESS_MAX_INDEX   (0x100 + S3G_STREAM_MAX)


static ssize_t s3g_proc_write(struct file *filp, const char __user *buf, size_t count, loff_t *offp)
{
    s3g_card_t *s3g = PDE_DATA(file_inode(filp));

    char mode = '0';
    int dump_index = 0;
    unsigned int ret = 0;
    char temp[20] = {0};
    unsigned int interval = 0, crtc = 0;
    unsigned long value = 0;

    if(count > 1 && count < 20)
    {
        ret = copy_from_user(&temp[0], buf, count);
        temp[count-1] = '\0';

        if(count == 2)
        {
            mode = temp[0];

            if(mode >= '0' && mode <= '9')
            {
                dump_index = (mode - '0');
            }
            else if((mode >= 'a' && mode < 'e'))
            {
                dump_index =  mode - 'a' + 0xa;
            }
            else if(mode == 'p')
            {
                dump_index = S3G_DUMP_POWER_STATUS_INDEX;
            }
            else if(mode == 'k')
            {
                dump_index = S3G_DUMP_CLIPS_INDEX;
            }
            else
            {
               s3g_error("echo error proc mode, should between 0 and d \n");
               return -EFAULT;
            }
        }
        else
        {
            value = simple_strtoul(&temp[1], NULL, 16);

            s3g_info("value %d \n", value);

            //input format should be: [T|t][0~9]
            if(temp[0] == 't' || temp[0] == 'T')
            {
                dump_index = S3G_DUMP_SENSOR_TEMP_INDEX;
                interval = value;
            }
            else if(temp[0] == 'f' || temp[0] == 'F')//set dvfs clamp
            {
                dump_index = S3G_FORCE_DVFS_INDEX;
            }
            else if(temp[0] == 'v' || temp[0] == 'V')//force video index
            {
                dump_index = S3G_FORCE_DVFS_VIDEO_INDEX;
            }
            else if(temp[0] == 'i' || temp[0] == 'I')
            {
                s3g_begin_end_isp_dvfs(0, value);
            }
            else if((temp[0] == 'c' || temp[0] == 'C') &&
                    (temp[2] == 's' || temp[2] == 'S'))
            {
                if(s3g_strlen(temp) != 4)
                {
                    s3g_info("invalid flag buffer dump parameter!\n");
                }

                crtc = simple_strtoul(&temp[1], NULL, 16);
                dump_index = simple_strtoul(&temp[3], NULL, 16);

                dump_index += S3G_DUMP_COMPRESS_PS_INDEX;
            }
            else
            {
               s3g_error("invalid index\n");
            }
         }
    }
    else
    {
        s3g_error("proc write pass too long param, should less then 20\n");
    }

    if (dump_index == S3G_DUMP_CLIPS_INDEX)
    {
        s3g_dump_clips(s3g);
    }
    else if((dump_index >=0 && dump_index <=5) || dump_index == S3G_DUMP_POWER_STATUS_INDEX)
    {
        krnl_dump_resource(s3g->adapter, dump_index, 0);
    }
    else if(dump_index > 5 && dump_index < 0xe )
    {
        krnl_save_force_dvfs_index(s3g->adapter, dump_index);
    }
    else if(dump_index == S3G_DUMP_SENSOR_TEMP_INDEX)
    {
        krnl_dump_sensor_temprature(s3g->adapter, interval);
    }
    else if(dump_index == S3G_FORCE_DVFS_INDEX)
    {
        krnl_force_dvfs_index(s3g->adapter, value);
    }
    else if(dump_index == S3G_FORCE_DVFS_VIDEO_INDEX)
    {
        krnl_force_dvfs_video_index(s3g->adapter, value);
    }
    else if((dump_index >= S3G_DUMP_COMPRESS_PS_INDEX)&&
            (dump_index <= S3G_DUMP_COMPRESS_MAX_INDEX))
    {
        krnl_dump_resource(s3g->adapter, dump_index, crtc);
    }

    return count;
}


static const struct file_operations s3g_proc_fops =
{
    .owner = THIS_MODULE,
    .read  = s3g_proc_read,
    .write = s3g_proc_write,
};
#endif

static  ssize_t  disp_dbg_level_get(struct device *dev, struct device_attribute *attr, char *buf)
{
    int  dbg_level = 0;
    void**  adapter = dev_get_drvdata(dev);
    dbg_level = krnl_get_dbg_level(*adapter);

    return sprintf(buf, "%x\n", dbg_level);
}

static  ssize_t  disp_dbg_level_set(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
    void**  adapter = dev_get_drvdata(dev);
    unsigned long dbg_level = 0;
    ssize_t  ret = 0;

    ret = kstrtoul(buf, 16, &dbg_level);       
    if(!ret)
    {
        krnl_set_dbg_level(*adapter, dbg_level);
        ret = size;
    }

    return ret;
}

static struct device_attribute dbg_level_attr = __ATTR(dbg_level, 0644, disp_dbg_level_get, disp_dbg_level_set);

void s3g_card_pre_init(s3g_card_t *s3g, void *pdev)
{
#ifdef CONFIG_FB
    s3g->pdev  = pdev;

    s3g->s3gfb_enable = s3g_fb;

    s3g_fb_detect_vesa(s3g);
#endif

    return;
}
int s3g_card_init(s3g_card_t *s3g, void *pdev)
{
    unsigned short command = 0;
    int ret = 0, primary;

    s3g->index = s3g - s3g_cards;

    spin_lock_init(&s3g->wait_list_lock);
    INIT_LIST_HEAD(&s3g->wait_list);

    s3g_init_bus_id(s3g);

    s3g_get_command_status16(pdev, &command);

    primary = (command & 0x01) ? 1 : 0;

    s3g_info("s3g_card_init: command=0x%x, primary=%x, vesafb_enable=%d, vga16fb_enable=%d\n",
            command, primary, s3g->vesafb_enable, s3g->vga16fb_enable);

    /* we can enable s3gfb primay and no fb driver enable by default */
    if((!primary || s3g->vesafb_enable || s3g->vga16fb_enable) && !s3g_run_on_qt)
    {
        s3g->s3gfb_enable = FALSE;
    }

    /* if fb not enable, we need reserve vmem for VGA, hardcode 256K for easy */
#ifdef S3G_PCIE_BUS
    if(primary && !(s3g->s3gfb_enable || s3g->vesafb_enable || s3g->vga16fb_enable))
    {
        s3g->reserved_vmem = 256 * 1024;
        s3g_info("reserve 256K vmem for VGA.\n");
    }
#endif

    s3g->freezable_patch = s3g_freezable_patch;

    s3g_info("s3g->freezable_patch : 0x%x.\n", s3g->freezable_patch);

    s3g->adapter = krnl_init_adapter(pdev, s3g->reserved_vmem, s3g->index, &s3g_export);

    if(s3g->adapter == NULL)
    {
        ret = -1;

        s3g_error("init adapter failed\n");
    }

    if(s3g->vesafb_enable)
    {
        krnl_notify_vesa_mode(s3g->adapter, &s3g->vesafb_mode);
    }

#ifdef S3G_PCIE_BUS
    s3g->msi_enable = 1;
#else
    s3g->msi_enable = 0;
#endif

#ifndef S3G_HW_NULL
    s3g_interrupt_init(s3g);
#endif

    s3g->lock = s3g_create_mutex();

    if(s3g->s3gfb_enable)
    {
#ifdef CONFIG_FB
        s3g_fb_probe(s3g, s3g_fb_mode, s3g_flip);
#endif
    }

    s3g->device = device_create(s3g_class, NULL, MKDEV(S3G_MAJOR, s3g->index), &(s3g->adapter), "%s%d", S3G_DEV_NAME, s3g->index);

    if(s3g->device == NULL)
    {
        ret = -1;
        s3g_error("device_create failed.\n");
    }

    if(device_create_file(s3g->device, &dbg_level_attr))
    {
        s3g_info("Create debug level control file failed!");
    }


#ifdef NO_PROC_CREATE_FUNC
    //create_proce_read_entry is not used any more in Linux 3.10 and above. use proc_create_data instead.
    create_proc_read_entry(S3G_PROC_NAME, 0, NULL, s3g_read_proc, s3g);
#else
    proc_create_data(S3G_PROC_NAME, 0, NULL, &s3g_proc_fops, s3g);
#endif

    s3g->debugfs_dev = s3g_debugfs_create_device(s3g);

    s3g_keyboard_init(s3g);
    s3g_rc_dev_init(s3g);
    s3g_sync_init(s3g);
    return ret;
}

int s3g_card_deinit(s3g_card_t *s3g)
{
    if(s3g->s3gfb_enable)
    {
#ifdef CONFIG_FB
        s3g_fb_remove(s3g);
#endif
    }

    device_destroy(s3g_class, MKDEV(S3G_MAJOR, s3g->index));

    s3g->device = NULL;

    krnl_wait_chip_idle(s3g->adapter);

#ifndef S3G_HW_NULL
    s3g_interrupt_deinit(s3g);
#endif
    krnl_deinit_adapter(s3g->adapter);

    s3g->len = 0;

    if(s3g->lock)
    {
        s3g_destroy_mutex(s3g->lock);
        s3g->lock = NULL;
    }

    if (s3g->sync_dev)
    {
        s3g_sync_deinit(s3g);
        s3g->sync_dev = NULL;
    }
    if(s3g->rc_dev)
    {
        s3g_rc_dev_deinit(s3g);
        s3g->rc_dev = NULL;
    }

    if(s3g->debugfs_dev)
    {
        s3g_debugfs_destroy_device((s3g_debugfs_device_t*)(s3g->debugfs_dev));
        s3g->debugfs_dev = NULL;
    }

    return 0;
}

int s3g_create_allocation(int card, s3g_create_allocation_t *create_data)
{
    s3g_card_t   *s3g = &s3g_cards[card];
    unsigned int ret  = 0;

    if (s3g->adapter)
    {
        ret = krnl_create_allocation(s3g->adapter, create_data);
    }

    return ret;
}

void s3g_destroy_allocation(int card, s3g_destroy_allocation_t *destroy_data)
{
    s3g_card_t *s3g = &s3g_cards[card];

    if (s3g->adapter)
    {
        krnl_destroy_allocation(s3g->adapter, destroy_data);
    }
}

int s3g_lock_allocation(int card, s3g_lock_allocation_t *lock_data)
{
    s3g_card_t  *s3g = &s3g_cards[card];
    unsigned int ret = 0;

    if (s3g->adapter)
    {
        ret = krnl_lock_allocation(s3g->adapter, lock_data, S3G_MEM_KERNEL);
    }

    return ret;
}

void s3g_unlock_allocation(int card, s3g_unlock_allocation_t *unlock_data)
{
    s3g_card_t *s3g = &s3g_cards[card];

    if (s3g->adapter)
    {
        krnl_unlock_allocation(s3g->adapter, unlock_data, S3G_MEM_KERNEL);
    }
}

void s3g_flush_allocation(int card, s3g_flush_allocation_t *flush_data)
{
    s3g_card_t *s3g = &s3g_cards[card];

    if (s3g->adapter)
    {
        krnl_flush_allocation(s3g->adapter, flush_data);
    }
}

void s3g_open_allocation(int card, s3g_open_allocation_t *open_data)
{
    s3g_card_t *s3g = &s3g_cards[card];

    if (s3g->adapter)
    {
        krnl_open_allocation(s3g->adapter, open_data);
    }
}

int s3g_create_device(int card, unsigned int *device)
{
    s3g_card_t  *s3g = &s3g_cards[card];
    unsigned int ret = 0;

    if (s3g->adapter)
    {
        ret = krnl_create_device(s3g->adapter, NULL, device);
    }

    return ret;
}

void s3g_destroy_device(int card, unsigned int device)
{
    s3g_card_t *s3g = &s3g_cards[card];

    if (s3g->adapter)
    {
        krnl_destroy_device(s3g->adapter, device);
    }
}

void s3g_begin_end_isp_dvfs(int card, unsigned int on)
{
    s3g_card_t *s3g = &s3g_cards[card];

    if (s3g->adapter)
    {
        krnl_begin_end_isp_dvfs(s3g->adapter, on);
    }
}

EXPORT_SYMBOL(s3g_create_allocation);
EXPORT_SYMBOL(s3g_destroy_allocation);
EXPORT_SYMBOL(s3g_lock_allocation);
EXPORT_SYMBOL(s3g_unlock_allocation);

EXPORT_SYMBOL(s3g_create_device);
EXPORT_SYMBOL(s3g_destroy_device);

EXPORT_SYMBOL(s3g_begin_end_isp_dvfs);
