/* 
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
* 
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "s3g_def.h"
#include "s3g_ioctl.h"
#include "os_interface.h"
#include "kernel_interface.h"
#include "s3g_screen.h"
#include "s3g_sync.h"
#include "s3g_buffer_object.h"
#include "s3g_debugfs.h"
/*****************************************************************************
**
** ioctl interface
**
*****************************************************************************/
int s3g_ioctl_addctx(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user           *argp = (void __user*)arg;
    s3g_ctx_t              ctx;

    if(s3g_copy_from_user(&ctx, argp, sizeof(ctx)))
    {
        return -1;
    }

    ctx.handle = 100;

    if(s3g_copy_to_user(argp, &ctx, sizeof(ctx)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_rmctx(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    return 0;
}

int s3g_ioctl_lock_ctx(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    //we clear freezable before lock thread, this value record the  thread real value to restore when unlock.
    int                   freezable = s3g_freezable();

    if (freezable)
    {
        if ((s3g->freezable_patch == 0) || (s3g->freezable_patch == 1))
        {
            s3g_clear_freezable();
        }
    }

    if (s3g_mutex_lock_killable(s3g->lock))
    {
        return -1;
    }

    priv->hold_lock = TRUE;
    priv->freezable = freezable;

    return 0;
}

int s3g_ioctl_unlock_ctx(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;

    priv->hold_lock = FALSE;

    s3g_mutex_unlock(s3g->lock);

    if(priv->freezable)
    {
        if (s3g->freezable_patch == 0)
        {
            s3g_set_freezable();
        }
        else if (s3g->freezable_patch == 1)
        {
            s3g_old_set_freezable();
        }
    }

    return 0;
}

int s3g_ioctl_get_busid(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;

    void __user           *argp = (void __user*)arg;
    char __user           *unique;
    s3g_unique_t          u;


    if(s3g_copy_from_user(&u, argp, sizeof(u)))
    {
        return -1;
    }

    if(u.unique_len >= s3g->len)
    {
        unique = ptr64_to_ptr(u.unique);
        if(s3g_copy_to_user(unique, s3g->busId, s3g->len))
        {
            return -1;
        }
    }

    u.unique_len = s3g->len;

    if(s3g_copy_to_user(argp, &u, sizeof(u)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_wait_func(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t    *s3g  = priv->card;
    void __user   *argp = (void __user*)arg;

    s3g_wait_func_t wait;

    if(s3g_copy_from_user(&wait, argp, sizeof(s3g_wait_func_t)))
    {
        return -1;
    }

    switch (wait.opcode)
    {
    case S3G_WAIT_ADD:
        s3g_add_wait(s3g, priv, wait.events);
        break;
    case S3G_WAIT_REMOVE:
        s3g_remove_wait(s3g, priv);
        break;
    case S3G_WAIT_GET_EVENT:
        s3g_get_event(s3g, priv, &wait);
        break;
    default:
        s3g_error("unknown wait opcode: %d.\n", wait.opcode);
        break;
    }

    if(s3g_copy_to_user(argp, &wait, sizeof(s3g_wait_func_t)))
    {
        return -1;
    }

    return 0;
}


int s3g_ioctl_get_display_info(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_display_info_t    display_info;

    if(s3g_copy_from_user(&display_info, argp, sizeof(s3g_display_info_t)))
    {
        return -1;
    }

    if(s3g_get_display_info(s3g, &display_info))
    {
        return -1;
    }

    if(s3g_copy_to_user(argp, &display_info, sizeof(display_info)))
    {
        return -1;
    }

    return 0;
}


int s3g_ioctl_get_iga_info(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;
    s3g_iga_info_t                iga_info;

    if(s3g_copy_from_user(&iga_info, argp, sizeof(s3g_iga_info_t)))
    {
        return -1;
    }

    krnl_get_iga_info(s3g->adapter, &iga_info);
    
    if(s3g_copy_to_user(argp, &iga_info, sizeof(iga_info)))
    {
        return -1;
    }

    return 0;
}

static int s3g_ioctl_clip_notify(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;
    s3g_clipnotify_t              clip;

    if(s3g_copy_from_user(&clip, argp, sizeof(clip)))
    {
        return -1;
    }

    s3g_clip_notify(s3g, &clip);

    return 0;
}

static int s3g_ioctl_video_drawable_damage(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;
    s3g_video_damage_t            damage;

    if(s3g_copy_from_user(&damage, argp, sizeof(s3g_video_damage_t)))
    {
        return -1;
    }

    s3g_video_drawable_damage(s3g, &damage);

    if(s3g_copy_to_user(argp, &damage, sizeof(s3g_video_damage_t)))
    {
        return -1;
    }

    return 0;
}

static int s3g_ioctl_add_drawable(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;
    s3g_drawable_t                drawable = {0, };

    if(s3g_copy_from_user(&drawable, argp, sizeof(s3g_drawable_t)))
    {
        return -1;
    }

    s3g_add_drawable(s3g, &drawable);

    if(s3g_copy_to_user(argp, &drawable, sizeof(s3g_drawable_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_remove_drawable(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;
    unsigned int                  drawable_id;

    if(s3g_copy_from_user(&drawable_id, argp, sizeof(drawable_id)))
    {
        return -1;
    }

    s3g_remove_drawable(s3g, drawable_id);

    return 0;
}

int s3g_ioctl_get_drawsize(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;

    s3g_get_drawsize_t            draw_rec;

    if(s3g_copy_from_user(&draw_rec, argp, sizeof(draw_rec)))
    {
        return -1;
    }

    s3g_get_drawable_info(s3g, &draw_rec);

    if(s3g_copy_to_user(argp, &draw_rec, sizeof(draw_rec)))
    {
        return -1;
    }

    return 0;
}

 int  s3g_ioctl_get_device_comb(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_dev_comb_t     device_comb;

    if(s3g_copy_from_user(&device_comb, argp, sizeof(s3g_dev_comb_t)))
    {
        return -1;
    }

    krnl_get_device_comb(s3g->adapter, &device_comb);

    if(s3g_copy_to_user(argp, &device_comb, sizeof(s3g_dev_comb_t)))
    {
        return -1;
    }

    return 0;    
}

int  s3g_ioctl_get_hpd_info(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_hpd_info_t         hpd_info;

    if(s3g_copy_from_user(&hpd_info, argp, sizeof(s3g_hpd_info_t)))
    {
        return -1;
    }

    krnl_get_hpd_info(s3g->adapter, &hpd_info);

    if(s3g_copy_to_user(argp, &hpd_info, sizeof(s3g_hpd_info_t)))
    {
        return -1;
    }

    return 0;    
}

int s3g_ioctl_query_display_caps(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_display_caps_t     display_caps;

    if(s3g_copy_from_user(&display_caps, argp, sizeof(display_caps)))
    {
        return -1;
    }

    krnl_query_display_caps(s3g->adapter, &display_caps);

    if(s3g_copy_to_user(argp, &display_caps, sizeof(display_caps)))
    {
        return -1;
    }

    return 0;
}

static int s3g_ioctl_query_output_caps(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_output_caps_t  output_caps;

    if(s3g_copy_from_user(&output_caps, argp, sizeof(s3g_output_caps_t)))
    {
        return -1;
    }    

    krnl_query_output_caps(s3g->adapter, &output_caps, 1);

    if(s3g_copy_to_user(argp, &output_caps, sizeof(output_caps)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_set_mode(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_mode_info_t       mode;

    if(s3g_copy_from_user(&mode, argp, sizeof(s3g_mode_info_t)))
    {
        return -1;
    }

    if(mode.crtc == 0)
    {
        s3g_memcpy(&s3g->tv_mode, &mode, sizeof(s3g_mode_info_t));
    }

    krnl_set_mode(s3g->adapter, &mode);

    return 0;
}

int s3g_ioctl_register_mhl_callback(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;

    krnl_register_mhl_callback(s3g->adapter);

    return 0;
}

int s3g_ioctl_get_hdmi_audio_format(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;

    s3g_hdmi_audio_formats         audio_formats = {0};

    if(krnl_get_hdmi_audio_format(s3g->adapter, &audio_formats))
    {
        return -1;
    }

    if(s3g_copy_to_user(argp, &audio_formats, sizeof(s3g_hdmi_audio_formats)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_cursor(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_set_cursor_t       cursor_set;

    if(s3g_copy_from_user(&cursor_set, argp, sizeof(s3g_set_cursor_t)))
    {
        return -1;
    }

    krnl_set_cursor(s3g->adapter, &cursor_set);

    return 0;
}

static int s3g_ioctl_gamma(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    int                   ret   = -1;

    s3g_gamma_info_t  gamma = {0};

    unsigned int* lut_buffer = s3g_malloc(256 * sizeof(unsigned int));

    if(!s3g_copy_from_user(&gamma, argp, sizeof(gamma)) 
       && !s3g_copy_from_user(lut_buffer, gamma.pLutBuf, 256 * sizeof(unsigned int)))
    {
        gamma.pLutBuf = lut_buffer;
    
        krnl_set_gamma(s3g->adapter, &gamma);

        ret = 0;
    }

    s3g_free(lut_buffer);

    return ret;
}

int s3g_ioctl_dpms(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_dpms_set_t        dpms_set;
    
    if(s3g_copy_from_user(&dpms_set, argp, sizeof(dpms_set)))
    {
        return -1;
    }

    krnl_set_dpms(s3g->adapter, &dpms_set);

    return 0;
}

int s3g_ioctl_get_mode_timing(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                    *s3g  = priv->card;
    void __user                   *argp = (void __user*)arg;

    s3g_mode_timing_info_t        timing_info;

    if(s3g_copy_from_user(&timing_info, argp, sizeof(timing_info)))
    {
        return -1;
    }

    if(krnl_get_mode_timing(s3g->adapter, &timing_info))
    {
        return -1;
    }

    if(s3g_copy_to_user(argp, &timing_info, sizeof(timing_info)))
    {
        return -1;
    }

    return 0;
}


int s3g_ioctl_2d_init(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user       *argp = (void __user*)arg;
    int               ret   = -1;

    s3g_2d_init_t     init;

    s3g_copy_from_user(&init, argp, sizeof(init));

    switch (init.func)
    {
    case S3G_INIT:
        s3g_set_server_info(priv->card, init.server_index, init.xinerama);

        init.device      = 
        priv->gpu_device = krnl_create_2ddevice(priv->card->adapter, priv->filp, init.client_type);

        priv->need_cleanup_drawable = 1;
        priv->server_index = init.server_index;

        ret = 0;
        break;

    case S3G_CLEANUP:
        ret = krnl_destroy_2ddevice(priv->card->adapter, priv->filp);
        priv->gpu_device = 0;
        break;

    default:
        break;

    }
    
    s3g_copy_to_user(argp, &init, sizeof(init));

    return ret;
}

static int s3g_ioctl_init_screen(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user       *argp = (void __user*)arg;

    s3g_2d_screen_t   screen;

    if(s3g_copy_from_user(&screen, argp, sizeof(s3g_2d_screen_t)))
    {
        return -1;
    }

    s3g_init_2d_screen(priv->card, priv, &screen);

    if(s3g_copy_to_user(argp, &screen, sizeof(s3g_2d_screen_t)))
    {
        return -1;
    }

    return 0;
}

static int s3g_ioctl_deinit_screen(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user       *argp = (void __user*)arg;

    s3g_2d_screen_t   screen;

    if(s3g_copy_from_user(&screen, argp, sizeof(s3g_2d_screen_t)))
    {
        return -1;
    }

    s3g_deinit_2d_screen(priv->card, &screen);

    return 0;
}

int s3g_ioctl_wait_chip_idle(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t  *s3g = priv->card;

    krnl_wait_chip_idle(s3g->adapter);

    return 0;
}

int _s3g_get_max_covered_iga(s3g_file_t *priv, unsigned int hwDrawableId);

#define MAX2(a,b)       ((a) > (b) ? (a) : (b))
#define MAX4(a,b,c,d)   MAX2(MAX2(a,b), MAX2(c,d))

#define MIN2(a,b)       ((a) > (b) ? (b) : (a))
#define MIN4(a,b,c,d)   MIN2(MIN2(a,b),MIN2(c,d))
int _s3g_get_max_covered_iga(s3g_file_t *priv, unsigned int hwDrawableId)
{
    s3g_card_t      *s3g  = priv->card;

    s3g_drawable_t          *drawable = NULL;
    s3g_iga_virtual_info_t  *vinfo = NULL;
    s3g_clipnotify_t        *clip;
    unsigned int   iga_index;
    int   coveredx = 0, coveredy = 0, covered_size = 0, max_covered = 0;
    int   max_covered_iga = -1;

    drawable = s3g_get_drawable(hwDrawableId);

    for(iga_index = 0; iga_index < MAX_CORE_CRTCS; iga_index++)
    {
        vinfo = s3g->iga_virtual_info + iga_index;
        if (vinfo->active)
        {
            if (max_covered_iga == -1)
                max_covered_iga = iga_index;

            clip = &(drawable->clipNotify[0]);

            coveredx = vinfo->width + clip->width - 
                MAX4(vinfo->virtualX, vinfo->virtualX + vinfo->width, clip->screen_x, clip->screen_x + clip->width) +
                MIN4(vinfo->virtualX, vinfo->virtualX + vinfo->width, clip->screen_x, clip->screen_x + clip->width);

            coveredy = vinfo->height + clip->height - 
                MAX4(vinfo->virtualY, vinfo->virtualY + vinfo->height, clip->y, clip->screen_y + clip->height) +
                MIN4(vinfo->virtualY, vinfo->virtualY + vinfo->height, clip->y, clip->screen_y + clip->height);

            if (coveredx > 0 && coveredy > 0)
            {
                covered_size = coveredx * coveredy;
                if (covered_size > max_covered)
                {
                    max_covered = covered_size;
                    max_covered_iga = iga_index;
                }
            }
        }
    }

    return max_covered_iga;
}

int s3g_ioctl_wait_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t  *s3g = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_wait_func_t  wait_func_para = {0};
    int   max_covered_iga = -1;
    
    if(s3g_copy_from_user(&wait_func_para, argp, sizeof(s3g_wait_func_t)))
    {
        return -1;
    }

    if(wait_func_para.events == S3G_EVENT_WAIT_VSYNC_WITH_DRAWABLEID)
    {
        max_covered_iga = _s3g_get_max_covered_iga(priv, wait_func_para.hwDrawableId);
        if(max_covered_iga != -1)
        {
            switch(max_covered_iga)
            {
                case 0:
                    wait_func_para.events = S3G_EVENT_VSYNC_0_INTERRUPT;
                    break;

                case 1:
                    wait_func_para.events = S3G_EVENT_VSYNC_1_INTERRUPT;
                    break;

                case 2:
                    wait_func_para.events = S3G_EVENT_VSYNC_2_INTERRUPT;
                    break;

                default:
                    break;
            }
        }
    }

    krnl_wait_event(s3g->adapter, &wait_func_para);
    
    return 0;
}

static int s3g_ioctl_escape_call(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                       *s3g  = priv->card;
    void __user                      *argp = (void __user*)arg;

    krnl_hint_t                      hint  = {0};
    s3g_escape_call_t                escape;
    int                              ret = 0;

    if(s3g_copy_from_user(&escape, argp, sizeof(s3g_escape_call_t)))
    {
        return -1;
    }

    hint.filp    = priv->filp;
    hint.hDevice = priv->gpu_device;


    ret = krnl_escape_call(s3g->adapter, &hint, &escape);

    if(hint.create_device)
    {
        priv->gpu_device = hint.hDevice;
        if(s3g->debugfs_dev)
        {
            s3g_debugfs_add_node(s3g->debugfs_dev, DEBUGFS_NODE_DEVICE, s3g_get_current_pid(), priv->gpu_device, DEBUGFS_ESCAPE_CREATE);
        }
    }
    else if(hint.destroy_device)
    {
        if(s3g->debugfs_dev)
        {
            s3g_debugfs_remove_node(s3g->debugfs_dev, priv->gpu_device);
        }

        priv->gpu_device = 0;
    }

    return ret;
}

static int s3g_ioctl_2d_draw(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                       *s3g  = priv->card;
    void __user                      *argp = (void __user*)arg;

    s3g_2d_draw_t                    draw_2d;

    if(s3g_copy_from_user(&draw_2d, argp, sizeof(s3g_2d_draw_t)))
    {
        return -1;
    }
    
    if( draw_2d.len > 0 )
    {
        draw_2d.kernel_mode_command = 0;
        krnl_kickoff_2d_draw(s3g->adapter, &draw_2d);
    }

    return 0;
}

int s3g_ioctl_acquire_aperture(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                       *s3g  = priv->card;
    void __user                      *argp = (void __user*)arg;
    s3g_acquire_aperture_t           acquire_aperture = {0};
    
    if(s3g_copy_from_user(&acquire_aperture, argp, sizeof(s3g_acquire_aperture_t)))
    {
        return -1;
    }

    krnl_acquire_aperture(s3g->adapter, &acquire_aperture);

    return 0;
 }

int s3g_ioctl_release_aperture(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                       *s3g  = priv->card;
    void __user                      *argp = (void __user*)arg;
    s3g_release_aperture_t           release_aperture = {0};
    
    if(s3g_copy_from_user(&release_aperture, argp, sizeof(s3g_release_aperture_t)))
    {
        return -1;
    }

    krnl_release_aperture(s3g->adapter, &release_aperture);

    return 0;
}

int s3g_ioctl_leave_vt(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g  = priv->card;

    s3g->enter_vt = FALSE;

    krnl_leave_s3g_graphics_mode(s3g->adapter, priv->filp);

#ifdef NEW_S3GFB
    if(s3g->info && s3g->s3gfb_pending_events)
    {
        unsigned int events = s3g->s3gfb_pending_events;

        s3g->s3gfb_pending_events = 0;

        s3gfb_handle_event(s3g, events);
    }
#endif

    return 0;
}

int s3g_ioctl_enter_vt(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g  = priv->card;

    s3g->enter_vt = TRUE;

    krnl_enter_s3g_graphics_mode(s3g->adapter, priv->filp);

    return 0;
}

int s3g_ioctl_check_device_combination(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                       *s3g  = priv->card;
    void __user                      *argp = (void __user*)arg;
    s3g_check_device_combination_t   device_combination;

    if(s3g_copy_from_user(&device_combination, argp, sizeof(device_combination)))
    {
        return -1;
    }

    if(krnl_check_device_combination(s3g->adapter, &device_combination))
    {
        return -1;
    }

    if(s3g_copy_to_user(argp, &device_combination, sizeof(device_combination)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_set_ps(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;
    framebuffer_t     *fb   = NULL;
    s3g_ps_info_t     ps;
    
    if(s3g_copy_from_user(&ps, argp, sizeof(s3g_ps_info_t)))
    {
        return -1;
    }

    fb = s3g_get_frame_buffer(s3g, ps.server_index, ps.screen_index);

    if(fb != NULL)
    {
        fb->rot_allocation[ps.crtc] = ps.shadow_surface;
    }

    krnl_set_ps(s3g->adapter, &ps);

    return 0;
}

int s3g_ioctl_flip(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;
    s3g_flip_info_t   flip;
    s3g_create_fence_fd_t create_fence = {0};

    if(s3g_copy_from_user(&flip, argp, sizeof(s3g_flip_info_t)))
    {
        return -1;
    }

    flip.retire_fence_fd = -1;
    flip.retire_fence_object = 0;

    if(flip.need_retire_fence && flip.hDevice)
    {
        create_fence.fence_value   = 1;
        create_fence.device        = flip.hDevice;

        if(s3g_sync_fence_create_fd(s3g->sync_dev, &create_fence) == 0)
        {
            flip.retire_fence_fd        = create_fence.fd;
            flip.retire_fence_object    = create_fence.sync_obj;
        }
    }

    krnl_flip(s3g->adapter, &flip);

    if (s3g_copy_to_user(argp, &flip, sizeof(s3g_flip_info_t)))
    {
        return -1;
    }
    return 0;
}

int s3g_ioctl_enum_sub_device(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int status = -1;

    if (krnl_enum_sub_device(priv->card->adapter) == S_OK)
    {
        status = 0;
    }

    return status;
}

int s3g_ioctl_create_allocation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    s3g_create_allocation_t create_data;
    int                     ret = 0;

    if(s3g_copy_from_user(&create_data, argp, sizeof(s3g_create_allocation_t)))
    {
        return -1;
    }

    if(create_data.need_fd)
    {
        //LeonHe:
        //s3g_buffer_oject_create() is not used in e2uma
        s3g_assert(0);

        //ret = s3g_buffer_object_create(s3g->adapter, &create_data);
        if(ret < 0)
        {
            return -1;
        }
    }
    else
    {
        ret = krnl_create_allocation(s3g->adapter, &create_data);
    }

    if(s3g_copy_to_user(argp, &create_data, sizeof(s3g_create_allocation_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_destroy_allocation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    s3g_destroy_allocation_t destroy_data;

    if(s3g_copy_from_user(&destroy_data, argp, sizeof(s3g_destroy_allocation_t)))
    {
        return -1;
    }

    krnl_destroy_allocation(s3g->adapter, &destroy_data);

    return 0;
}

static int s3g_ioctl_lock_allocation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    s3g_lock_allocation_t lock_data;
    int                   ret;

    if(s3g_copy_from_user(&lock_data, argp, sizeof(s3g_lock_allocation_t)))
    {
        return -1;
    }

    ret = krnl_lock_allocation(s3g->adapter, &lock_data, S3G_MEM_USER);

    if(s3g_copy_to_user(argp, &lock_data, sizeof(s3g_lock_allocation_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_unlock_allocation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    s3g_unlock_allocation_t unlock_data;

    if(s3g_copy_from_user(&unlock_data, argp, sizeof(s3g_unlock_allocation_t)))
    {
        return -1;
    }

    krnl_unlock_allocation(s3g->adapter, &unlock_data, S3G_MEM_USER);

    if(s3g_copy_to_user(argp, &unlock_data, sizeof(s3g_unlock_allocation_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_flush_allocation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    s3g_flush_allocation_t flush_data;

    if(s3g_copy_from_user(&flush_data, argp, sizeof(s3g_flush_allocation_t)))
    {
        return -1;
    }

    krnl_flush_allocation(s3g->adapter, &flush_data);

    return 0;
}

static int s3g_ioctl_ipc_operation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t              *s3g  = priv->card;
    void __user             *argp = (void __user*)arg;
    int                     ret=0;

    s3g_ipc_operation_t     ipc_operation;

    if(s3g_copy_from_user(&ipc_operation, argp, sizeof(s3g_ipc_operation_t)))
    {
        return -1;
    }

    ret = s3g_ipc_operation(s3g, priv, &ipc_operation);

    return ret;
}

int s3g_ioctl_set_interactive(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g = priv->card;
    void __user       *argp = (void __user*)arg;
    int    active = 0;

    if(s3g_copy_from_user(&active, argp, sizeof(int)))
    {
        return -1;
    }

    krnl_set_interactive(s3g->adapter, active);
    return  0;
}

int s3g_ioctl_open_allocation(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    s3g_open_allocation_t open_data;
    int                   ret;

    if(s3g_copy_from_user(&open_data, argp, sizeof(s3g_open_allocation_t)))
    {
        return -1;
    }

    ret = krnl_open_allocation(s3g->adapter, &open_data);

    if(s3g_copy_to_user(argp, &open_data, sizeof(s3g_open_allocation_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_wait_allocation_idle(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t                 *s3g  = priv->card;
    void __user                *argp = (void __user*)arg;

    s3g_wait_allocation_idle_t  wait_allocation;

    if(s3g_copy_from_user(&wait_allocation, argp, sizeof(s3g_wait_allocation_idle_t)))
    {
        return -1;
    }

    krnl_wait_allocation_idle(s3g->adapter, &wait_allocation);

    return 0;
}

int s3g_ioctl_video_get_hw_caps(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t  *s3g  = priv->card;
    void __user *argp = (void __user*)arg;

    unsigned char caps = krnl_get_video_hardware_capability(s3g->adapter);

    if(s3g_copy_to_user(argp, &caps, sizeof(unsigned char)))
    {
        return -1;
    }

    return 0;
}

static int s3g_ioctl_get_address_config(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t  *s3g  = priv->card;
    void __user *argp = (void __user*)arg;

    s3g_get_address_config_t address_config;

    krnl_get_address_config(s3g->adapter, &address_config);

    if(s3g_copy_to_user(argp, &address_config, sizeof(s3g_get_address_config_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_query_info(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t  *s3g  = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_query_info_t info;
    int ret = 0;
    
    if(s3g_copy_from_user(&info, argp, sizeof(s3g_query_info_t)))
    {
        return -1;
    }

    ret = krnl_query_info(s3g->adapter, &info);

    if(s3g_copy_to_user(argp, &info, sizeof(s3g_query_info_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_create_stream(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;
    s3g_create_stream_t  create_stream;
    int ret = 0;

    if(s3g_copy_from_user(&create_stream, argp, sizeof(s3g_create_stream_t)))
    {
        return -1;
    }

    ret =  krnl_create_stream(s3g->adapter, &create_stream);

    if(s3g_copy_to_user(argp, &create_stream, sizeof(s3g_create_stream_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_destroy_stream(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;
    s3g_destroy_stream_t  destroy_stream;
    int ret = 0;

    if(s3g_copy_from_user(&destroy_stream, argp, sizeof(s3g_destroy_stream_t)))
    {
        return -1;
    }

    krnl_destroy_stream(s3g->adapter, &destroy_stream);

    if(s3g_copy_to_user(argp, &destroy_stream, sizeof(s3g_destroy_stream_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_display_frame(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g          = priv->card;
    void __user           *argp         = (void __user*)arg;
    s3g_display_frame_t   display_frame;
    s3g_create_fence_fd_t create_fence  = {0};

    int ret = 0;

    if(s3g_copy_from_user(&display_frame, argp, sizeof(s3g_display_frame_t)))
    {
        return -1;
    }

    //active_stream_mask should be subset of acquired_stream_mask
    if(display_frame.active_stream_mask != (display_frame.active_stream_mask & display_frame.acquired_stream_mask))
    {
        s3g_error("s3gDisplayFrame invalid mask:%x-%x\n", display_frame.active_stream_mask, display_frame.acquired_stream_mask);
        return -1;
    }

    display_frame.retire_fence_fd     = -1;
    display_frame.retire_fence_object = 0;

    if(display_frame.need_retire_fence && display_frame.hDevice)
    {
        create_fence.fence_value = S3G_DISPLAY_RETIRE_SIGNAL_VALUE;
        create_fence.device      = display_frame.hDevice;
        create_fence.context     = display_frame.hContext;

        if(s3g_sync_fence_create_fd(s3g->sync_dev, &create_fence) == 0)
        {
            display_frame.retire_fence_fd     = create_fence.fd;
            display_frame.retire_fence_object = create_fence.sync_obj;
        }
    }

    display_frame.release_fence_fd     = -1;
    display_frame.release_fence_object = 0;

    if(display_frame.need_release_fence && display_frame.hDevice)
    {
        create_fence.fence_value = S3G_DISPLAY_RELEASE_SIGNAL_VALUE;
        create_fence.device      = display_frame.hDevice;

        if(s3g_sync_fence_create_fd(s3g->sync_dev, &create_fence) == 0)
        {
            display_frame.release_fence_fd     = create_fence.fd;
            display_frame.release_fence_object = create_fence.sync_obj;
        }
    }

    ret =  krnl_display_frame(s3g->adapter, &display_frame);

    if(s3g_copy_to_user(argp, &display_frame, sizeof(s3g_display_frame_t)))
    {
        return -1;
    }

 //   krnl_free_vesafb_tempbuffer(s3g->adapter);

    return ret;
}

int s3g_ioctl_cec_ctl(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;
    s3g_cec_ctl_t   cec_ctl;
    int ret = 0;

    if(s3g_copy_from_user(&cec_ctl, argp, sizeof(s3g_cec_ctl_t)))
    {
        return -1;
    }

    s3g_info("s3g ioctl cec ctl\n");

    ret = krnl_cec_ctl(s3g->adapter, &cec_ctl);

    if(s3g_copy_to_user(argp, &cec_ctl, sizeof(s3g_cec_ctl_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_present(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;

    s3g_drawable_t  *drawable = NULL;
    s3g_present_t   present;

    unsigned int  last = 0, scrn_index, status = 0;

    if(s3g_copy_from_user(&present, argp, sizeof(s3g_present_t)))
    {
        return -1;
    }

    if(present.ignore_drawable)
    {
        s3g_clipnotify_t clip = {0, };
        s3g_clip_rect_t stack_rects[8];
        s3g_clip_rect_t *clip_rect = stack_rects;

        if (present.num_cliprects > 0)
        {
            if (present.num_cliprects > sizeof(stack_rects)/sizeof(stack_rects[0]))
            {
                clip_rect = s3g_malloc(sizeof(*clip_rect) * present.num_cliprects);
            }
            s3g_copy_from_user(clip_rect, present.cliprects, present.num_cliprects * sizeof(*clip_rect));

            clip.hAllocation = present.dst_allocation;
            clip.clipRects = clip_rect;
            clip.numClipRects = present.num_cliprects;
        }
        else
        {
            framebuffer_t *fb = s3g_get_frame_buffer(s3g, present.server_index, present.screen_index);

            if (present.dst_allocation)
            {
                /* this case means blit src_allocation to dst_allocation */
                clip.hAllocation = present.dst_allocation;
            }
            else
            {
                /* this case means flip fb->hAllocation to diu */
                clip.hAllocation = fb->hAllocation;
            }
            clip_rect->x1 = 0;
            clip_rect->y1 = 0;
            clip_rect->x2 = fb->width;
            clip_rect->y2 = fb->height;
            clip.clipRects = clip_rect;
            clip.numClipRects = 1;
        }

        status = krnl_present(s3g->adapter, &present, &clip);

        if (clip_rect != stack_rects)
        {
            s3g_free(clip_rect);
        }

        return status;
    }

    drawable  = s3g_get_drawable(present.drawable_id);

    for(scrn_index = 0; scrn_index < MAX_SCREENS; scrn_index++)
    {
        if(drawable->clipNotify[scrn_index].hAllocation 
           && last != drawable->clipNotify[scrn_index].hAllocation)
        {
            framebuffer_t *fb = s3g_get_frame_buffer(priv->card, present.server_index, present.screen_index);

            if (fb && fb->adapter == s3g->adapter)
            {
                status = krnl_present(s3g->adapter, &present, &drawable->clipNotify[scrn_index]);
                last   = drawable->clipNotify[scrn_index].hAllocation;
            }

            if(status != 0)
            {
                break;
            }
        }
    }

    return status;
}

int s3g_ioctl_get_drawable_surface(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;

    s3g_get_drawable_surface_t   get_surface;

    unsigned int    hAllocation = 0;

    if(s3g_copy_from_user(&get_surface, argp, sizeof(s3g_get_drawable_surface_t)))
    {
        return -1;
    }

    /* drawable_id 0 means get primary surface */
    if(get_surface.drawable_id == 0)
    {
        framebuffer_t *fb = s3g_get_frame_buffer(s3g, get_surface.server_index, get_surface.screen_index);

        hAllocation = fb->hAllocation;
    }
    else
    {
        s3g_drawable_t *drawable = s3g_get_drawable(get_surface.drawable_id);

        hAllocation = drawable->clipNotify[get_surface.screen_index].hAllocation;
    }

    get_surface.surface = hAllocation;

    if(s3g_copy_to_user(argp, &get_surface, sizeof(s3g_get_drawable_surface_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_get_vsync_cnt(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;

    s3g_get_vsync_cnt_t     get_vsync;
    int   max_covered_iga = -1;

    if(s3g_copy_from_user(&get_vsync, argp, sizeof(s3g_get_vsync_cnt_t)))
    {
        return -1;
    }

    max_covered_iga = _s3g_get_max_covered_iga(priv, get_vsync.drawable_id);

    if(max_covered_iga == -1)
    {
        return -1;
    }

    get_vsync.vsync_cnt = s3g->v_sync_count[max_covered_iga];

    if(s3g_copy_to_user(argp, &get_vsync, sizeof(s3g_get_vsync_cnt_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_create_device(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_create_device_t    create_device = {0};
    int                    ret  = 0;

    s3g_assert(priv->gpu_device == 0);

    ret = krnl_create_device(s3g->adapter, priv->filp, &create_device.device);

    if(s3g_copy_to_user(argp, &create_device, sizeof(s3g_create_device_t)))
    {
        return -1;
    }

    priv->gpu_device = create_device.device;

    if(s3g->debugfs_dev)
    {
        s3g_debugfs_add_node(s3g->debugfs_dev, DEBUGFS_NODE_DEVICE, s3g_get_current_pid(), priv->gpu_device, DEBUGFS_S3G_CREATE);
    }

    return ret;
}

int s3g_ioctl_destroy_device(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g = priv->card;
    void __user       *argp = (void __user*)arg;

    unsigned int       device = 0;


    if(s3g_copy_from_user(&device, argp, sizeof(unsigned int)))
    {
        return -1;
    }

    s3g_info("destroy device %x.\n", device);

    if(s3g->debugfs_dev)
    {
        s3g_debugfs_remove_node(s3g->debugfs_dev, priv->gpu_device);
    }

    krnl_destroy_device(s3g->adapter, device);

    priv->gpu_device = 0;

    return 0;
}

int s3g_ioctl_begin_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g = priv->card;
    void __user  *argp = (void __user*)arg;
    s3g_begin_perf_event_t begin_perf_event = {0};
    int          ret = 0;

    if (s3g_copy_from_user(&begin_perf_event, argp, sizeof(s3g_begin_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_begin_perf_event(s3g->adapter, &begin_perf_event);

    return ret;
}

int s3g_ioctl_begin_miu_dump_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t *s3g = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_begin_miu_dump_perf_event_t begin_miu_dump = {0};
    int ret = 0;

    if(s3g_copy_from_user(&begin_miu_dump, argp, sizeof(s3g_begin_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_begin_miu_dump_perf_event(s3g->adapter, &begin_miu_dump);

    return ret;
}

int s3g_ioctl_end_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g = priv->card;
    void __user  *argp = (void __user*)arg;
    s3g_end_perf_event_t end_perf_event = {0};
    int          ret = 0;

    if (s3g_copy_from_user(&end_perf_event, argp, sizeof(s3g_end_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_end_perf_event(s3g->adapter, &end_perf_event);

    return ret;
}

int s3g_ioctl_end_miu_dump_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t *s3g = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_end_miu_dump_perf_event_t end_miu_dump = {0};
    int ret = 0;

    if(s3g_copy_from_user(&end_miu_dump, argp, sizeof(s3g_end_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_end_miu_dump_perf_event(s3g->adapter, &end_miu_dump);

    return ret;
}


static int s3g_ioctl_set_miu_reg_list_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t *s3g = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_miu_reg_list_perf_event_t miu_reg_list = {0};
    int ret = 0;

    if(s3g_copy_from_user(&miu_reg_list, argp, sizeof(s3g_miu_reg_list_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_set_miu_reg_list_perf_event(s3g->adapter, &miu_reg_list);

    return ret;
}

static int s3g_ioctl_get_miu_dump_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t *s3g = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_get_miu_dump_perf_event_t get_miu_dump = {0};
    int ret = 0;

    if(s3g_copy_from_user(&get_miu_dump, argp, sizeof(s3g_get_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_get_miu_dump_perf_event(s3g->adapter, &get_miu_dump);

    if (ret == 0)
    {
        s3g_copy_to_user(argp, &get_miu_dump, sizeof(s3g_get_miu_dump_perf_event_t));
    }

    return ret;
}

static int s3g_ioctl_direct_get_miu_dump_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t *s3g = priv->card;
    void __user *argp = (void __user*)arg;
    s3g_direct_get_miu_dump_perf_event_t direct_get_dump = {0};
    int ret = 0;

    if(s3g_copy_from_user(&direct_get_dump, argp, sizeof(s3g_direct_get_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_direct_get_miu_dump_perf_event(s3g->adapter, &direct_get_dump);

    if(ret == 0)
    {
        s3g_copy_to_user(argp, &direct_get_dump, sizeof(s3g_direct_get_miu_dump_perf_event_t));
    }

    return ret;
}

static int s3g_ioctl_get_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g = priv->card;
    void __user  *argp = (void __user*)arg;
    s3g_get_perf_event_t get_perf_event = {0};
    int          ret = 0;

    if (s3g_copy_from_user(&get_perf_event, argp, sizeof(s3g_get_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_get_perf_event(s3g->adapter, &get_perf_event);

    if (ret == 0)
    {
        s3g_copy_to_user(argp, &get_perf_event, sizeof(s3g_get_perf_event_t));
    }

    return ret;
}

int s3g_ioctl_send_perf_event(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g = priv->card;
    void __user  *argp = (void __user*)arg;
    s3g_perf_event_t perf_event;
    int          ret = 0;

    if (s3g_copy_from_user(&perf_event, argp, sizeof(s3g_perf_event_header_t)))
    {
        return -1;
    }

    if (perf_event.header.size > sizeof(s3g_perf_event_header_t))
    {
        if (s3g_copy_from_user(&perf_event, argp, perf_event.header.size))
        {
            return -1;
        }
    }

    ret = krnl_send_perf_event(s3g->adapter, &perf_event);

    return ret;
}


int s3g_ioctl_get_perf_status(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g = priv->card;
    void __user  *argp = (void __user*)arg;
    s3g_perf_status_t perf_status = {0};
    int ret = 0;

    ret = krnl_get_perf_status(s3g->adapter, &perf_status);

    if (ret == 0)
    {
        s3g_copy_to_user(argp, &perf_status, sizeof(s3g_perf_status_t));
    }

    return ret;
}

static int s3g_ioctl_create_context(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t          *s3g  = priv->card;
    void __user         *argp = (void __user*)arg;

    s3g_create_context_t create_context = {0};
    int                  ret  = 0;

    if(s3g_copy_from_user(&create_context, argp, sizeof(s3g_create_context_t)))
    {
        return -1;
    }

    ret = krnl_create_context(s3g->adapter, &create_context, S3G_MEM_USER);

    if (ret != S_OK)
    {
        return -1;
    }

    if(s3g_copy_to_user(argp, &create_context, sizeof(s3g_create_context_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_destroy_context(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t          *s3g  = priv->card;
    void __user         *argp = (void __user*)arg;

    s3g_destroy_context_t destroy_context = {0};
    int                   ret  = 0;

    if(s3g_copy_from_user(&destroy_context, argp, sizeof(s3g_destroy_context_t)))
    {
        return -1;
    }

    ret = krnl_destroy_context(s3g->adapter, &destroy_context);

    if(s3g_copy_to_user(argp, &destroy_context, sizeof(s3g_destroy_context_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_create_di_context(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t          *s3g  = priv->card;
    void __user         *argp = (void __user*)arg;

    s3g_create_di_context_t create_context;
    int                  ret  = 0;

    if(s3g_copy_from_user(&create_context, argp, sizeof(s3g_create_di_context_t)))
    {
        return -1;
    }

    ret = krnl_create_di_context(s3g->adapter, &create_context);

    if(s3g_copy_to_user(argp, &create_context, sizeof(s3g_create_di_context_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_destroy_di_context(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t          *s3g  = priv->card;
    void __user         *argp = (void __user*)arg;

    s3g_destroy_di_context_t destroy_context;
    int                      ret  = 0;

    if(s3g_copy_from_user(&destroy_context, argp, sizeof(s3g_destroy_di_context_t)))
    {
        return -1;
    }

    ret = krnl_destroy_di_context(s3g->adapter, &destroy_context);

    if(s3g_copy_to_user(argp, &destroy_context, sizeof(s3g_destroy_di_context_t)))
    {
        return -1;
    }

    return ret;
}

static int s3g_ioctl_render(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g    = priv->card;
    void __user  *argp   = (void __user*)arg;

    s3g_render_t  render = {0};
    int           ret    = 0;

    if(s3g_copy_from_user(&render, argp, sizeof(s3g_render_t)))
    {
        return -1;
    }

    ret = krnl_render(s3g->adapter, &render);

    if(s3g_copy_to_user(argp, &render, sizeof(s3g_render_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_query_chip_id(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;
    unsigned int      ret = 0;

    s3g_chip_id_t     chip_id;
    s3g_query_info_t  query;

    if(s3g_copy_from_user(&chip_id, argp, sizeof(s3g_chip_id_t)))
    {
        return -1;
    }

    query.type = S3G_QUERY_VENDOR_ID;

    krnl_query_info(s3g->adapter, &query);

    chip_id.Vendor_ID = query.value;

    query.type = S3G_QUERY_DEVICE_ID;

    krnl_query_info(s3g->adapter, &query);

    chip_id.Device_ID = query.value >> 8;

    query.type = S3G_QUERY_REVISION_ID;

    krnl_query_info(s3g->adapter, &query);

    chip_id.Revision_ID = query.value;

    if(s3g_copy_to_user(argp, &chip_id, sizeof(s3g_chip_id_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_hdcp_op(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;
    unsigned int      ret;

    s3g_hdcp_op_t     hdcp_op;

    if(s3g_copy_from_user(&hdcp_op, argp, sizeof(s3g_hdcp_op_t)))
    {
        return -1;
    }

    ret = krnl_hdcp_op(s3g->adapter, priv, &hdcp_op);

    if(s3g_copy_to_user(argp, &hdcp_op, sizeof(s3g_hdcp_op_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_create_fence_sync_object(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    s3g_create_fence_sync_object_t  create = {0};

    if(s3g_copy_from_user(&create, argp, sizeof(s3g_create_fence_sync_object_t)))
    {
        return -1;
    }

    result = krnl_create_fence_sync_object(s3g->adapter, &create, FALSE);

    if(s3g_copy_to_user(argp, &create, sizeof(s3g_create_fence_sync_object_t)))
    {
        return -1;
    }

    return result;
}

int s3g_ioctl_destroy_fence_sync_object(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    s3g_destroy_fence_sync_object_t  destroy = {0};

    if(s3g_copy_from_user(&destroy, argp, sizeof(s3g_destroy_fence_sync_object_t)))
    {
        return -1;
    }

    result = krnl_destroy_fence_sync_object(s3g->adapter, &destroy);

    return result;
}

int s3g_ioctl_wait_fence_sync_object(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    s3g_wait_fence_sync_object_t  wait = {0};

    if(s3g_copy_from_user(&wait, argp, sizeof(s3g_wait_fence_sync_object_t)))
    {
        return -1;
    }

    result = krnl_wait_fence_sync_object(s3g->adapter, &wait);

    if(s3g_copy_to_user(argp, &wait, sizeof(s3g_wait_fence_sync_object_t)))
    {
        return -1;
    }

    return result;
}

int s3g_ioctl_fence_value(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    s3g_fence_value_t  value = {0};

    if(s3g_copy_from_user(&value, argp, sizeof(s3g_fence_value_t)))
    {
        return -1;
    }

    result = krnl_fence_value(s3g->adapter, &value);

    if(s3g_copy_to_user(argp, &value, sizeof(s3g_fence_value_t)))
    {
        return -1;
    }

    return result;
}

int s3g_ioctl_create_fence_fd(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t   *s3g    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    s3g_create_fence_fd_t value = {0};

    if(s3g_copy_from_user(&value, argp, sizeof(s3g_create_fence_fd_t)))
    {
        return -1;
    }

    result = s3g_sync_fence_create_fd(s3g->sync_dev, &value);

    if(s3g_copy_to_user(argp, &value, sizeof(s3g_create_fence_fd_t)))
    {
        return -1;
    }

    return result;

}

int s3g_ioctl_flip_limit(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_flip_limit_t    flipvalue;

    if(s3g_copy_from_user(&flipvalue, argp, sizeof(s3g_flip_limit_t)))
    {
        return -1;
    }

    krnl_flip_limit(s3g->adapter, &flipvalue);

   return 0;
}

int s3g_ioctl_wait_context_idle(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g  = priv->card;
    void __user       *argp = (void __user*)arg;

    unsigned int context = 0;

    if(s3g_copy_from_user(&context, argp, sizeof(unsigned int)))
    {
        return -1;
    }

    krnl_wait_context_idle(s3g->adapter, context);

    return 0;
}

static int s3g_ioctl_send_dsi_cmd(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;
    s3g_dsi_cmd_t dsi_cmd;
    int ret = 0;

    if(s3g_copy_from_user(&dsi_cmd, argp, sizeof(s3g_dsi_cmd_t)))
    {
        return -1;
    }

    ret = krnl_send_dsi_cmd(s3g->adapter, &dsi_cmd);

    if(s3g_copy_to_user(argp, &dsi_cmd, sizeof(s3g_dsi_cmd_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_mhl_cbus_send_command(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;
    s3g_cbus_cmd_t cbus_cmd;
    int ret = 0;

    if(s3g_copy_from_user(&cbus_cmd, argp, sizeof(s3g_cbus_cmd_t)))
    {
        return -1;
    }

    ret = krnl_mhl_cbus_send_command(s3g->adapter, &cbus_cmd);

    if(s3g_copy_to_user(argp, &cbus_cmd, sizeof(s3g_cbus_cmd_t)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_set_power_perf_mode(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t            *s3g  = priv->card;
    void __user           *argp = (void __user*)arg;

    s3g_power_perf_mode_info_t       mode;

    if(s3g_copy_from_user(&mode, argp, sizeof(s3g_power_perf_mode_info_t)))
    {
        return -1;
    }

    krnl_set_power_perf_mode(s3g->adapter, &mode);

    return 0;
}

int s3g_ioctl_gfx_stm(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t      *s3g  = priv->card;
    void __user     *argp = (void __user*)arg;
    int             is_resume = 0;
    int ret = 0;

    if(s3g_copy_from_user(&is_resume, argp, sizeof(int)))
    {
        return -1;
    }

    if(is_resume)
    {
        s3g_interrupt_reinit(s3g);
        ret = krnl_restore_state(s3g->adapter);
    }
    else
    {
        krnl_save_state(s3g->adapter, 1);
        s3g_interrupt_deinit(s3g);
    }

    if(s3g_copy_to_user(argp, &is_resume, sizeof(is_resume)))
    {
        return -1;
    }

    return ret;
}

int s3g_ioctl_fd_to_handle(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int             ret;
    void __user     *argp = (void __user*)arg;
    s3g_handle_fd_t hfd;

    if(s3g_copy_from_user(&hfd, argp, sizeof(s3g_handle_fd_t)))
    {
        return -1;
    }

    hfd.handle = 0;

    ret = s3g_bo_fd_to_handle(priv, &hfd);
    if (ret)
    {
        return ret;
    }

    if(s3g_copy_to_user(argp, &hfd, sizeof(s3g_handle_fd_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_handle_to_fd(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int             ret = 0;
    void __user     *argp = (void __user*)arg;
    s3g_handle_fd_t hfd;

    if(s3g_copy_from_user(&hfd, argp, sizeof(s3g_handle_fd_t)))
    {
        return -1;
    }

    hfd.fd = -1;

    ret = s3g_bo_handle_to_fd(priv, &hfd);

    if (ret)
    {
        return ret;
    }

    if(s3g_copy_to_user(argp, &hfd, sizeof(s3g_handle_fd_t)))
    {
        return -1;
    }
    return 0;
}

int s3g_ioctl_get_ust_msc(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g = priv->card;
    void __user       *argp = (void __user*)arg;
    s3g_ust_msc_t     um = {0, };

    if(s3g_copy_from_user(&um, argp, sizeof(s3g_ust_msc_t)))
    {
        return -1;
    }

    if (um.crtc < 0 || um.crtc >= sizeof(s3g->v_sync_count) / sizeof(s3g->v_sync_count[0]))
    {
        return -1;
    }

    um.msc = s3g->v_sync_count[um.crtc];
    um.ust = s3g_do_div(s3g->v_sync_timestamp[um.crtc] , 1000);

    if(s3g_copy_to_user(argp, &um, sizeof(s3g_ust_msc_t)))
    {
        return -1;
    }

    return 0;
}

int s3g_ioctl_iga_virtual_info(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    s3g_card_t        *s3g = priv->card;
    void __user       *argp = (void __user*)arg;
    s3g_iga_virtual_info_t info = {0, };

    if (s3g_copy_from_user(&info, argp, sizeof(info)))
    {
        return -1;
    }

    if (info.iga_index < 0 || info.iga_index >= sizeof(s3g->iga_virtual_info) / sizeof(s3g->iga_virtual_info[0]))
    {
        return -1;
    }

    s3g_memcpy(s3g->iga_virtual_info + info.iga_index, &info, sizeof(info));

    return 0;
}

int s3g_ioctl_get_allocation_state(s3g_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int               ret = -1;
    s3g_card_t        *s3g = priv->card;
    void __user       *argp = (void __user*)arg;
    s3g_get_allocation_state_t state = {0, };

    if (s3g_copy_from_user(&state, argp, sizeof(state)))
    {
        return -1;
    }

    ret = krnl_get_allocation_state(s3g->adapter, &state);

    if (ret == 0)
    {
        ret = s3g_copy_to_user(argp, &state, sizeof(state));
    }

    return ret;
}


s3g_ioctl_t s3g_ioctls[] =
{
    [S3G_IOCTL_NR(S3G_IOCTL_GET_BUSID)]                = s3g_ioctl_get_busid,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_DISPLAYINFO)]          = s3g_ioctl_get_display_info,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_IGAINFO)]              = s3g_ioctl_get_iga_info,
    [S3G_IOCTL_NR(S3G_IOCTL_ADD_CTX)]                  = s3g_ioctl_addctx,
    [S3G_IOCTL_NR(S3G_IOCTL_RM_CTX)]                   = s3g_ioctl_rmctx,
    [S3G_IOCTL_NR(S3G_IOCTL_LOCK)]                     = s3g_ioctl_lock_ctx,
    [S3G_IOCTL_NR(S3G_IOCTL_UNLOCK)]                   = s3g_ioctl_unlock_ctx,
    [S3G_IOCTL_NR(S3G_IOCTL_CLIP_NOTIFY)]              = s3g_ioctl_clip_notify,
    [S3G_IOCTL_NR(S3G_IOCTL_ADD_DRAW)]                 = s3g_ioctl_add_drawable,
    [S3G_IOCTL_NR(S3G_IOCTL_RM_DRAW)]                  = s3g_ioctl_remove_drawable,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_DRAWSIZE)]             = s3g_ioctl_get_drawsize,
    [S3G_IOCTL_NR(S3G_IOCTL_QUERY_DISPLAY_CAPS)]       = s3g_ioctl_query_display_caps,
    [S3G_IOCTL_NR(S3G_IOCTL_QUERY_OUTPUT_CAPS)]        = s3g_ioctl_query_output_caps,
    [S3G_IOCTL_NR(S3G_IOCTL_MODE_SET)]                 = s3g_ioctl_set_mode,
    [S3G_IOCTL_NR(S3G_IOCTL_CURSOR)]                   = s3g_ioctl_cursor,
    [S3G_IOCTL_NR(S3G_IOCTL_GAMMA)]                    = s3g_ioctl_gamma,
    [S3G_IOCTL_NR(S3G_IOCTL_DPMS)]                     = s3g_ioctl_dpms,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_MODE_TIMING)]          = s3g_ioctl_get_mode_timing,
    [S3G_IOCTL_NR(S3G_IOCTL_2D_INIT)]                  = s3g_ioctl_2d_init,
    [S3G_IOCTL_NR(S3G_IOCTL_INIT_SCREEN)]              = s3g_ioctl_init_screen,
    [S3G_IOCTL_NR(S3G_IOCTL_DEINIT_SCREEN)]            = s3g_ioctl_deinit_screen,    
    [S3G_IOCTL_NR(S3G_IOCTL_WAIT_CHIP_IDLE)]           = s3g_ioctl_wait_chip_idle,
    [S3G_IOCTL_NR(S3G_IOCTL_ESCAPE_CALL)]              = s3g_ioctl_escape_call,
    [S3G_IOCTL_NR(S3G_IOCTL_2D_DRAW)]                  = s3g_ioctl_2d_draw,
    [S3G_IOCTL_NR(S3G_IOCTL_ACQUIRE_APERTURE)]         = s3g_ioctl_acquire_aperture,
    [S3G_IOCTL_NR(S3G_IOCTL_RELEASE_APERTURE)]         = s3g_ioctl_release_aperture,
    [S3G_IOCTL_NR(S3G_IOCTL_LEAVE_VT)]                 = s3g_ioctl_leave_vt,
    [S3G_IOCTL_NR(S3G_IOCTL_ENTER_VT)]                 = s3g_ioctl_enter_vt,
    [S3G_IOCTL_NR(S3G_IOCTL_CHECK_DEVICE_COMBINATION)] = s3g_ioctl_check_device_combination,
    [S3G_IOCTL_NR(S3G_IOCTL_SET_PS)]                   = s3g_ioctl_set_ps,
    [S3G_IOCTL_NR(S3G_IOCTL_ENUM_SUB_DEVICE)]          = s3g_ioctl_enum_sub_device,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_ALLOCATION)]        = s3g_ioctl_create_allocation,
    [S3G_IOCTL_NR(S3G_IOCTL_OPEN_ALLOCATION)]          = s3g_ioctl_open_allocation,
    [S3G_IOCTL_NR(S3G_IOCTL_DESTROY_ALLOCATION)]       = s3g_ioctl_destroy_allocation,
    [S3G_IOCTL_NR(S3G_IOCTL_LOCK_ALLOCATION)]          = s3g_ioctl_lock_allocation,
    [S3G_IOCTL_NR(S3G_IOCTL_UNLOCK_ALLOCATION)]        = s3g_ioctl_unlock_allocation,
    [S3G_IOCTL_NR(S3G_IOCTL_FLUSH_ALLOCATION)]         = s3g_ioctl_flush_allocation,
    [S3G_IOCTL_NR(S3G_IOCTL_WAIT_ALLOCATION_IDLE)]     = s3g_ioctl_wait_allocation_idle,
    [S3G_IOCTL_NR(S3G_IOCTL_VIDEO_GET_HW_CAPS)]        = s3g_ioctl_video_get_hw_caps,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_ADDRESS_CONFIG)]       = s3g_ioctl_get_address_config,
    [S3G_IOCTL_NR(S3G_IOCTL_QUERY_INFO)]               = s3g_ioctl_query_info,
    [S3G_IOCTL_NR(S3G_IOCTL_WAIT_FUNC)]                = s3g_ioctl_wait_func,
    [S3G_IOCTL_NR(S3G_IOCTL_FLIP)]                     = s3g_ioctl_flip,
    [S3G_IOCTL_NR(S3G_IOCTL_CEC_CTL)]                  = s3g_ioctl_cec_ctl,
    [S3G_IOCTL_NR(S3G_IOCTL_PRESENT)]                  = s3g_ioctl_present,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_DRAWABLE_SURFACE)]     = s3g_ioctl_get_drawable_surface,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_VSYNC_CNT)]            = s3g_ioctl_get_vsync_cnt,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_DEVICE)]            = s3g_ioctl_create_device,
    [S3G_IOCTL_NR(S3G_IOCTL_DESTROY_DEVICE)]           = s3g_ioctl_destroy_device,
    [S3G_IOCTL_NR(S3G_IOCTL_BEGIN_PERF_EVENT)]         = s3g_ioctl_begin_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_END_PERF_EVENT)]           = s3g_ioctl_end_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_PERF_EVENT)]           = s3g_ioctl_get_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_SEND_PERF_EVENT)]          = s3g_ioctl_send_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_PERF_STATUS)]          = s3g_ioctl_get_perf_status,
    [S3G_IOCTL_NR(S3G_IOCTL_BEGIN_MIU_DUMP_PERF_EVENT)] = s3g_ioctl_begin_miu_dump_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_END_MIU_DUMP_PERF_EVENT)]  = s3g_ioctl_end_miu_dump_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_SET_MIU_REG_LIST_PERF_EVENT)] = s3g_ioctl_set_miu_reg_list_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_MIU_DUMP_PERF_EVENT)]  = s3g_ioctl_get_miu_dump_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_DIRECT_GET_MIU_DUMP_PERF_EVENT)] = s3g_ioctl_direct_get_miu_dump_perf_event,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_CONTEXT)]           = s3g_ioctl_create_context,
    [S3G_IOCTL_NR(S3G_IOCTL_DESTROY_CONTEXT)]          = s3g_ioctl_destroy_context,
    [S3G_IOCTL_NR(S3G_IOCTL_WAIT_CONTEXT_IDLE)]        = s3g_ioctl_wait_context_idle,
    [S3G_IOCTL_NR(S3G_IOCTL_RENDER)]                   = s3g_ioctl_render,
    [S3G_IOCTL_NR(S3G_IOCTL_QUERY_CHIP_ID)]            = s3g_ioctl_query_chip_id,
    [S3G_IOCTL_NR(S3G_IOCTL_HDCP_OP)]                  = s3g_ioctl_hdcp_op,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_FENCE_SYNC_OBJECT)] = s3g_ioctl_create_fence_sync_object,
    [S3G_IOCTL_NR(S3G_IOCTL_DESTROY_FENCE_SYNC_OBJECT)]= s3g_ioctl_destroy_fence_sync_object,
    [S3G_IOCTL_NR(S3G_IOCTL_WAIT_FENCE_SYNC_OBJECT)]   = s3g_ioctl_wait_fence_sync_object,
    [S3G_IOCTL_NR(S3G_IOCTL_FENCE_VALUE)]              = s3g_ioctl_fence_value,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_FENCE_FD)]          = s3g_ioctl_create_fence_fd,
    [S3G_IOCTL_NR(S3G_IOCTL_SET_FLIP_COUNTER)]         = s3g_ioctl_flip_limit,
    [S3G_IOCTL_NR(S3G_IOCTL_IPC_OPERATION)]            = s3g_ioctl_ipc_operation,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_STREAM)]            = s3g_ioctl_create_stream,
    [S3G_IOCTL_NR(S3G_IOCTL_DESTROY_STREAM)]           = s3g_ioctl_destroy_stream,
    [S3G_IOCTL_NR(S3G_IOCTL_DISPLAY_FRAME)]            = s3g_ioctl_display_frame,
    [S3G_IOCTL_NR(S3G_IOCTL_SET_INTERACTIVE)]          = s3g_ioctl_set_interactive,
    [S3G_IOCTL_NR(S3G_IOCTL_SEND_DSI_CMD)]             = s3g_ioctl_send_dsi_cmd,
    [S3G_IOCTL_NR(S3G_IOCTL_MHL_CBUS_SEND_CMD)]        = s3g_ioctl_mhl_cbus_send_command,
    [S3G_IOCTL_NR(S3G_IOCTL_CREATE_DI_CONTEXT)]        = s3g_ioctl_create_di_context,
    [S3G_IOCTL_NR(S3G_IOCTL_DESTROY_DI_CONTEXT)]       = s3g_ioctl_destroy_di_context,
    [S3G_IOCTL_NR(S3G_IOCTL_REGISTER_MHL_CALLBACK)]    = s3g_ioctl_register_mhl_callback,
    [S3G_IOCTL_NR(S3G_IOCTL_SET_POWER_PERF_MODE)]      = s3g_ioctl_set_power_perf_mode,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_DEVICE_COMB)]          = s3g_ioctl_get_device_comb,
    [S3G_IOCTL_NR(S3G_IOCTL_GFX_STM)]                  = s3g_ioctl_gfx_stm,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_HDMI_AUDIO_FORMAT)]    = s3g_ioctl_get_hdmi_audio_format,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_HPD_INFO)]             = s3g_ioctl_get_hpd_info,
    [S3G_IOCTL_NR(S3G_IOCTL_HANDLE_TO_FD)]             = s3g_ioctl_handle_to_fd,
    [S3G_IOCTL_NR(S3G_IOCTL_FD_TO_HANDLE)]             = s3g_ioctl_fd_to_handle,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_UST_MSC)]              = s3g_ioctl_get_ust_msc,
    [S3G_IOCTL_NR(S3G_IOCTL_GET_ALLOCATION_STATE)]     = s3g_ioctl_get_allocation_state,
    [S3G_IOCTL_NR(S3G_IOCTL_WAIT_EVENT)]               = s3g_ioctl_wait_event,
    [S3G_IOCTL_NR(S3G_IOCTL_VIDEO_DAMAGE)]             = s3g_ioctl_video_drawable_damage,
    [S3G_IOCTL_NR(S3G_IOCTL_IGA_VIRTUAL_INFO)]         = s3g_ioctl_iga_virtual_info,
};




